from unittest.mock import Mock

import pytest

from mode.utils.compat import isatty, want_bytes, want_str


@pytest.mark.parametrize(
    "input,expected",
    [
        ("foo", b"foo"),
        (b"foo", b"foo"),
    ],
)
def test_want_bytes(input, expected):
    assert want_bytes(input) == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (b"foo", "foo"),
        ("foo", "foo"),
    ],
)
def test_want_str(input, expected):
    assert want_str(input) == expected


def test_isatty():
    m = Mock()
    m.isatty.return_value = True
    assert isatty(m)
    m.isatty.side_effect = AttributeError()
    assert not isatty(m)



"""Worker - Starts services from the command-line.

Workers add signal handling, logging, and other things
required to start and manage services in a process environment.
"""
import asyncio
import logging as _logging
import os
import reprlib
import signal
import sys
import traceback
import typing
from contextlib import contextmanager, suppress
from logging import Handler, Logger
from typing import (
    IO,
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from .services import Service
from .types import ServiceT
from .utils import logging
from .utils.futures import all_tasks, maybe_cancel
from .utils.imports import symbol_by_name
from .utils.times import Seconds
from .utils.typing import NoReturn

if typing.TYPE_CHECKING:
    from .debug import BlockingDetector
else:

    class BlockingDetector:
        ...  # noqa


__all__ = ["Worker"]

logger = logging.get_logger(__name__)

EX_OK = getattr(os, "EX_OK", 0)
EX_FAILURE = 1
EX_OSERR = getattr(os, "EX_OSERR", 71)
BLOCK_DETECTOR = "mode.debug:BlockingDetector"


class _TupleAsListRepr(reprlib.Repr):
    def repr_tuple(self, x: Tuple, level: int) -> str:
        return self.repr_list(cast(list, x), level)


# this repr formats tuples as if they are lists.
_repr = _TupleAsListRepr().repr  # noqa: E305


@contextmanager
def exiting(*, print_exception: bool = False, file: IO = sys.stderr) -> Iterator[None]:
    try:
        yield
    except MemoryError:
        sys.stderr.write("Out of memory!")
        sys.exit(EX_OSERR)
    except Exception as exc:
        if print_exception:
            print(f"Command raised exception: {exc!r}", file=file)
            traceback.print_tb(exc.__traceback__, file=file)
        sys.exit(EX_FAILURE)
    sys.exit(EX_OK)


class Worker(Service):
    """Start mode service from the command-line."""

    BLOCK_DETECTOR: ClassVar[str] = BLOCK_DETECTOR

    stdout: IO
    stderr: IO
    debug: bool
    quiet: bool
    blocking_timeout: Seconds
    logging_config: Optional[Dict]
    loglevel: Optional[Union[str, int]]
    logfile: Optional[Union[str, IO]]
    console_port: int
    loghandlers: List[Handler]
    redirect_stdouts: bool
    redirect_stdouts_level: int

    services: Iterable[ServiceT]

    _blocking_detector: Optional[BlockingDetector] = None
    _starting_fut: Optional[asyncio.Future] = None

    # signals can be called multiple times,
    # so when stopped by signal we record the time to make sure
    # we don't start the process multiple times.
    _signal_stop_time: Optional[float] = None
    _signal_stop_future: Optional[asyncio.Future] = None

    def __init__(
        self,
        *services: ServiceT,
        debug: bool = False,
        quiet: bool = False,
        logging_config: Dict = None,
        loglevel: Union[str, int] = None,
        logfile: Union[str, IO] = None,
        redirect_stdouts: bool = True,
        redirect_stdouts_level: logging.Severity = None,
        stdout: Optional[IO] = sys.stdout,
        stderr: Optional[IO] = sys.stderr,
        console_port: int = 50101,
        loghandlers: List[Handler] = None,
        blocking_timeout: Seconds = 10.0,
        loop: asyncio.AbstractEventLoop = None,
        override_logging: bool = True,
        daemon: bool = True,
        **kwargs: Any,
    ) -> None:
        self.services = services
        self.debug = debug
        self.quiet = quiet
        self.logging_config = logging_config
        self.loglevel = loglevel
        self.logfile = logfile
        self.loghandlers = loghandlers or []
        self.redirect_stdouts = redirect_stdouts
        self.redirect_stdouts_level = logging.level_number(
            redirect_stdouts_level or "WARN"
        )
        self.override_logging = override_logging
        self.stdout = sys.stdout if stdout is None else stdout
        self.stderr = sys.stderr if stderr is None else stderr
        self.console_port = console_port
        self.blocking_timeout = blocking_timeout
        self.daemon = daemon
        super().__init__(loop=loop, **kwargs)

        if self.services:
            for service in self.services:
                service.beacon.reattach(self.beacon)
                assert service.beacon.root is self.beacon

    def say(self, msg: str) -> None:
        """Write message to standard out."""
        self._say(msg)

    def carp(self, msg: str) -> None:
        """Write warning to standard err."""
        self._say(msg, file=self.stderr)

    def _say(
        self, msg: str, file: Optional[IO] = None, end: str = "\n", **kwargs: Any
    ) -> None:
        if file is None:
            file = self.stdout
        if not self.quiet:
            print(msg, file=file, end=end, **kwargs)  # noqa: T003

    def on_init_dependencies(self) -> Iterable[ServiceT]:
        return self.services

    async def on_first_start(self) -> None:
        await self.default_on_first_start()

    async def default_on_first_start(self) -> None:
        if self.override_logging:
            self._setup_logging()
        await self.on_execute()
        if self.debug:
            await self._add_monitor()
        self.install_signal_handlers()

    async def on_execute(self) -> None:
        ...

    def _setup_logging(self) -> None:
        _loglevel: int = 0
        try:
            _loglevel = logging.setup_logging(
                loglevel=self.loglevel,
                logfile=self.logfile,
                logging_config=self.logging_config,
                loghandlers=self.loghandlers,
            )
        except Exception as exc:
            try:
                self.stderr.write(f"CANNOT SETUP LOGGING: {exc!r} from ")
                import traceback

                traceback.print_stack(file=self.stderr)
            except Exception:  # noqa: S110
                pass
            raise
        self.on_setup_root_logger(_logging.root, _loglevel)
        if self.redirect_stdouts:
            self._redirect_stdouts()

    def _redirect_stdouts(self) -> None:
        self.add_context(logging.redirect_stdouts(severity=self.redirect_stdouts_level))

    def on_setup_root_logger(self, logger: Logger, level: int) -> None:
        ...

    async def maybe_start_blockdetection(self) -> None:
        if self.debug:
            await self.blocking_detector.maybe_start()

    def install_signal_handlers(self) -> None:
        if sys.platform == "win32":
            self._install_signal_handlers_windows()
        else:
            self._install_signal_handlers_unix()

    def _install_signal_handlers_windows(self) -> None:
        signal.signal(signal.SIGTERM, self._on_win_sigterm)

    def _install_signal_handlers_unix(self) -> None:
        self.loop.add_signal_handler(signal.SIGINT, self._on_sigint)
        self.loop.add_signal_handler(signal.SIGTERM, self._on_sigterm)
        self.loop.add_signal_handler(signal.SIGUSR1, self._on_sigusr1)
        self.loop.add_signal_handler(signal.SIGUSR2, self._on_sigusr2)

    def _on_sigint(self) -> None:
        self.carp("-INT- -INT- -INT- -INT- -INT- -INT-")
        self._schedule_shutdown(signal.SIGINT)

    def _on_sigterm(self) -> None:
        self._schedule_shutdown(signal.SIGTERM)

    def _on_win_sigterm(self, signum: int, frame: Any) -> None:
        self._schedule_shutdown(signal.SIGTERM)

    def _on_sigusr1(self) -> None:
        self.add_future(self._cry())

    def _on_sigusr2(self) -> None:
        self._enter_debugger()

    async def _cry(self) -> None:
        logging.cry(file=self.stderr)

    def _enter_debugger(self) -> None:
        self.carp("Starting debugger...")
        import pdb  # noqa: T100

        pdb.set_trace()  # noqa: T100

    def _schedule_shutdown(self, signal: signal.Signals) -> None:
        if not self._signal_stop_time:
            self._signal_stop_time = self.loop.time()
            self._signal_stop_future = asyncio.ensure_future(
                self._stop_on_signal(signal), loop=self.loop
            )

    async def _stop_on_signal(self, signal: signal.Signals) -> None:
        self.log.info("Signal received: %s (%s)", signal, signal.value)
        await self.stop()
        maybe_cancel(self._starting_fut)

    def execute_from_commandline(self) -> NoReturn:
        self._starting_fut = None
        with exiting(file=self.stderr):
            try:
                self._starting_fut = asyncio.ensure_future(
                    self.start(),
                    loop=self.loop,
                )
                self.loop.run_until_complete(self._starting_fut)
            except asyncio.CancelledError:
                pass
            except MemoryError:
                raise
            except Exception as exc:
                self.log.exception("Error: %r", exc)
                raise
            finally:
                maybe_cancel(self._starting_fut)
                self.on_worker_shutdown()
                self.stop_and_shutdown()
        # for mypy NoReturn
        raise SystemExit(EX_OK)  # pragma: no cover

    def on_worker_shutdown(self) -> None:
        ...

    def stop_and_shutdown(self) -> None:
        if self._signal_stop_future and not self._signal_stop_future.done():
            self.loop.run_until_complete(self._signal_stop_future)
        elif not self._stopped.is_set():
            self.loop.run_until_complete(self.stop())
        self._shutdown_loop()

    def _shutdown_loop(self) -> None:
        # Gather futures created by us.
        self.log.info("Gathering service tasks...")
        with suppress(asyncio.CancelledError):
            self.loop.run_until_complete(self._gather_futures())
        # Gather absolutely all asyncio futures.
        self.log.info("Gathering all futures...")
        self._gather_all()
        try:
            # Wait until loop is fully stopped.
            while self.loop.is_running():
                self.log.info("Waiting for event loop to shutdown...")
                self.loop.stop()
                self.loop.run_until_complete(asyncio.sleep(1.0))
        except BaseException as exc:
            self.log.exception("Got exception while waiting: %r", exc)
        finally:
            # Then close the loop.
            fut = asyncio.ensure_future(self._sentinel_task(), loop=self.loop)
            self.loop.run_until_complete(fut)
            self.loop.stop()
            self.log.info("Closing event loop")
            self.loop.close()
            if self.crash_reason:
                self.log.critical(
                    "We experienced a crash! Reraising original exception..."
                )
                raise self.crash_reason from self.crash_reason

    async def _sentinel_task(self) -> None:
        await asyncio.sleep(1.0)

    def _gather_all(self) -> None:
        # sleeps for at most 10 * 0.1s
        for _ in range(10):
            if not len(all_tasks(loop=self.loop)):
                break
            self.loop.run_until_complete(asyncio.sleep(0.1))
        for task in all_tasks(loop=self.loop):
            task.cancel()

    async def on_started(self) -> None:
        if self.daemon:
            await self.wait_until_stopped()

    async def _add_monitor(self) -> Any:
        try:
            import aiomonitor
        except ImportError:
            self.log.warning("Cannot start console: aiomonitor is not installed")
        else:
            monitor = aiomonitor.start_monitor(
                port=self.console_port,
                loop=self.loop,
            )
            self.add_context(monitor)

    def _repr_info(self) -> str:
        return _repr(self.services)

    @property
    def blocking_detector(self) -> BlockingDetector:
        if self._blocking_detector is None:
            BlockDetector: Callable = symbol_by_name(self.BLOCK_DETECTOR)
            self._blocking_detector = BlockDetector(
                self.blocking_timeout,
                beacon=self.beacon,
                loop=self.loop,
            )
        return cast(BlockingDetector, self._blocking_detector)



# -*- coding: utf-8 -*-
import sys
from contextlib import suppress

from sphinx_celery import conf

sys.path.append(".")

extensions = []

globals().update(
    conf.build_config(
        "mode",
        __file__,
        project="Mode",
        # version_dev='2.0',
        # version_stable='1.4',
        canonical_url="http://mode-streaming.readthedocs.io",
        webdomain="",
        github_project="faust-streaming/mode",
        copyright="2017-2020",
        html_logo="images/logo.png",
        html_favicon="images/favicon.ico",
        html_prepend_sidebars=[],
        include_intersphinx={"python", "sphinx"},
        extra_extensions=[
            "sphinx.ext.napoleon",
            "sphinx_autodoc_annotation",
            "alabaster",
        ],
        extra_intersphinx_mapping={},
        # django_settings='testproj.settings',
        # from pathlib import Path
        # path_additions=[Path.cwd().parent / 'testproj']
        apicheck_ignore_modules=[
            "mode.loop.eventlet",
            "mode.loop.gevent",
            "mode.loop.uvloop",
            "mode.loop._gevent_loop",
            "mode.utils",
            "mode.utils._py37_contextlib",
            "mode.utils.graphs.formatter",
            "mode.utils.graphs.graph",
            "mode.utils.types",
        ],
    )
)

html_theme = "alabaster"
html_sidebars = {}
templates_path = ["_templates"]

autodoc_member_order = "bysource"

pygments_style = "sphinx"

# This option is deprecated and raises an error.
with suppress(NameError):
    del html_use_smartypants  # noqa

extensions.remove("sphinx.ext.viewcode")



"""Backport implementation of Python 3.7 contextlib."""
import abc
import sys
import types
from collections import deque
from contextlib import AbstractContextManager
from functools import wraps
from typing import (
    Any,
    Awaitable,
    Callable,
    ContextManager,
    Dict,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

import _collections_abc

from .typing import AsyncContextManager, Deque

__all__ = [
    "AbstractAsyncContextManager",
    "AsyncExitStack",
    "ExitStack",
    "asynccontextmanager",
    "nullcontext",
]

AsyncCallable = Callable[..., Awaitable]
PushArg = Union[ContextManager, Callable]
AsyncPushArg = Union[AsyncContextManager, AsyncCallable]


class AbstractAsyncContextManager(abc.ABC):
    """An abstract base class for asynchronous context managers."""

    async def __aenter__(self) -> "AbstractAsyncContextManager":
        """Return `self` upon entering the runtime context."""
        return self

    @abc.abstractmethod
    async def __aexit__(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        traceback: types.TracebackType,
    ) -> None:
        """Raise any exception triggered within the runtime context."""
        return None

    @classmethod
    def __subclasshook__(cls, C: Type) -> bool:
        if cls is AbstractAsyncContextManager:
            return cast(
                bool, _collections_abc._check_methods(C, "__aenter__", "__aexit__")
            )
        return cast(bool, NotImplemented)


class _AsyncGeneratorContextManager(AbstractAsyncContextManager):
    """Helper for :func:`asynccontextmanager`."""

    # this __init__ is taken from contextlib._GeneratorContextManagerBase
    # in CPython3.7.

    def __init__(self, func: Callable, args: Tuple, kwds: Dict) -> None:
        self.gen = func(*args, **kwds)
        self.func, self.args, self.kwds = func, args, kwds
        # Issue 19330: ensure context manager instances have good docstrings
        doc = getattr(func, "__doc__", None)
        if doc is None:
            doc = type(self).__doc__
        self.__doc__ = doc
        # Unfortunately, this still doesn't provide good help output when
        # inspecting the created context manager instances, since pydoc
        # currently bypasses the instance docstring and shows the docstring
        # for the class instead.
        # See http://bugs.python.org/issue19404 for more details.

    # rest of class is from contextlib._AsyncGeneratorContextManager

    async def __aenter__(self) -> Any:
        try:
            return await self.gen.__anext__()
        except StopAsyncIteration:
            raise RuntimeError("generator didn't yield") from None

    async def __aexit__(
        self,
        typ: Optional[Type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[types.TracebackType],
    ) -> None:
        if typ is None:
            try:
                await self.gen.__anext__()
            except StopAsyncIteration:
                return
            else:
                raise RuntimeError("generator didn't stop")
        else:
            if value is None:
                value = typ()
            # See _GeneratorContextManager.__exit__ for comments on subtleties
            # in this implementation
            try:
                await self.gen.athrow(typ, value, traceback)
                raise RuntimeError("generator didn't stop after throw()")
            except StopAsyncIteration as exc:
                return exc is not value  # type: ignore
            except RuntimeError as exc:
                if exc is value:
                    return False  # type: ignore
                # Avoid suppressing if a StopIteration exception
                # was passed to throw() and later wrapped into a RuntimeError
                # (see PEP 479 for sync generators; async generators also
                # have this behavior). But do this only if the exception
                # wrapped by the RuntimeError is actully Stop(Async)Iteration
                # (see issue29692).
                if isinstance(value, (StopIteration, StopAsyncIteration)):
                    if exc.__cause__ is value:
                        return False  # type: ignore
                raise
            except BaseException as exc:
                if exc is not value:
                    raise


def asynccontextmanager(func: Callable) -> Callable[..., AsyncContextManager]:
    """None."""

    @wraps(func)
    def helper(*args: Any, **kwds: Any) -> _AsyncGeneratorContextManager:
        return _AsyncGeneratorContextManager(func, args, kwds)

    return helper


class _BaseExitStack:
    """A base class for ExitStack and AsyncExitStack."""

    _exit_callbacks: Deque[Tuple[bool, Callable]]

    @staticmethod
    def _create_exit_wrapper(cm: ContextManager, cm_exit: Callable) -> Callable:
        def _exit_wrapper(
            exc_type: Type, exc: BaseException, tb: types.TracebackType
        ) -> Any:
            return cm_exit(cm, exc_type, exc, tb)

        return _exit_wrapper

    @staticmethod
    def _create_cb_wrapper(callback: Callable, *args: Any, **kwds: Any) -> Callable:
        def _exit_wrapper(
            exc_type: Type, exc: BaseException, tb: types.TracebackType
        ) -> None:
            callback(*args, **kwds)

        return _exit_wrapper

    def __init__(self) -> None:
        self._exit_callbacks = deque()

    def pop_all(self) -> "_BaseExitStack":
        """Preserve the context stack by transferring it to a new instance."""
        new_stack = type(self)()
        new_stack._exit_callbacks = self._exit_callbacks
        self._exit_callbacks = deque()
        return new_stack

    def push(self, exit: PushArg) -> PushArg:
        """Register a callback with the standard __exit__ method signature.

        Can suppress exceptions the same way __exit__ method can.
        Also accepts any object with an __exit__ method (registering a call
        to the method instead of the object itself).
        """
        # We use an unbound method rather than a bound method to follow
        # the standard lookup behaviour for special methods.
        _cb_type = type(exit)

        try:
            exit_method = _cb_type.__exit__  # type: ignore
        except AttributeError:
            # Not a context manager, so assume it's a callable.
            self._push_exit_callback(exit)  # type: ignore
        else:
            self._push_cm_exit(exit, exit_method)  # type: ignore
        return exit  # Allow use as a decorator.

    def enter_context(self, cm: ContextManager) -> Any:
        """Enters the supplied context manager.

        If successful, also pushes its __exit__ method as a callback and
        returns the result of the __enter__ method.
        """
        # We look up the special methods on the type to match the with
        # statement.
        _cm_type = type(cm)
        _exit = _cm_type.__exit__
        result = _cm_type.__enter__(cm)
        self._push_cm_exit(cm, _exit)
        return result

    def callback(self, callback: Callable, *args: Any, **kwds: Any) -> Callable:
        """Register an arbitrary callback and arguments.

        Cannot suppress exceptions.
        """
        _exit_wrapper = self._create_cb_wrapper(callback, *args, **kwds)

        # We changed the signature, so using @wraps is not appropriate, but
        # setting __wrapped__ may still help with introspection.
        _exit_wrapper.__wrapped__ = callback  # type: ignore
        self._push_exit_callback(_exit_wrapper)
        return callback  # Allow use as a decorator

    def _push_cm_exit(self, cm: ContextManager, cm_exit: Callable) -> None:
        # Helper to correctly register callbacks to __exit__ methods.
        _exit_wrapper = self._create_exit_wrapper(cm, cm_exit)
        _exit_wrapper.__self__ = cm  # type: ignore
        self._push_exit_callback(_exit_wrapper, True)

    def _push_exit_callback(self, callback: Callable, is_sync: bool = True) -> None:
        self._exit_callbacks.append((is_sync, callback))


# Inspired by discussions on http://bugs.python.org/issue13585
class ExitStack(_BaseExitStack, AbstractContextManager):
    """Context manager for dynamic management of a stack of exit callbacks.

    For example:
        with ExitStack() as stack:
            files = [stack.enter_context(open(fname)) for fname in filenames]
            # All opened files will automatically be closed at the end of
            # the with statement, even if attempts to open files later
            # in the list raise an exception.
    """

    def __enter__(self) -> "ExitStack":
        return self

    def __exit__(self, *exc_details: Any) -> None:
        received_exc = exc_details[0] is not None

        # We manipulate the exception state so it behaves as though
        # we were actually nesting multiple with statements
        frame_exc = sys.exc_info()[1]

        def _fix_exception_context(
            new_exc: BaseException, old_exc: BaseException
        ) -> None:
            # Context may not be correct, so find the end of the chain
            while 1:
                exc_context = new_exc.__context__
                if exc_context is old_exc:
                    # Context is already set correctly (see issue 20317)
                    return
                if exc_context is None or exc_context is frame_exc:
                    break
                new_exc = exc_context
            # Change the end of the chain to point to the exception
            # we expect it to reference
            new_exc.__context__ = old_exc

        # Callbacks are invoked in LIFO order to match the behaviour of
        # nested context managers
        suppressed_exc = False
        pending_raise = False
        while self._exit_callbacks:
            is_sync, cb = self._exit_callbacks.pop()
            assert is_sync
            try:
                if cb(*exc_details):
                    suppressed_exc = True
                    pending_raise = False
                    exc_details = (None, None, None)
            except:  # noqa
                new_exc_details = sys.exc_info()
                # simulate the stack of exceptions by setting the context
                _fix_exception_context(
                    new_exc_details[1], exc_details[1]  # type: ignore
                )
                pending_raise = True
                exc_details = new_exc_details
        if pending_raise:
            try:
                # bare "raise exc_details[1]" replaces our carefully
                # set-up context
                fixed_ctx = exc_details[1].__context__
                raise exc_details[1]
            except BaseException:
                exc_details[1].__context__ = fixed_ctx
                raise
        return received_exc and suppressed_exc  # type: ignore

    def close(self) -> None:
        """Immediately unwind the context stack."""
        self.__exit__(None, None, None)


# Inspired by discussions on https://bugs.python.org/issue29302
class AsyncExitStack(_BaseExitStack, AbstractAsyncContextManager):
    """Async :class:`ExitStack`.

    Context manager for dynamic management of a stack of exit callbacks.

    Example:
            >>> async with AsyncExitStack() as stack:
            ...    connections = [
            ...        await stack.enter_async_context(get_connection())
            ...        for i in range(5)
            ...    ]
            >>>    # All opened connections will automatically be released at
            >>>    # the end of the async with statement, even if attempts to
            >>>    # open a connection later in the list raise an exception.
    """

    @staticmethod
    def _create_async_exit_wrapper(
        cm: AsyncContextManager, cm_exit: AsyncCallable
    ) -> AsyncCallable:
        async def _exit_wrapper(
            exc_type: Type, exc: BaseException, tb: types.TracebackType
        ) -> Any:
            return await cm_exit(cm, exc_type, exc, tb)

        return _exit_wrapper

    @staticmethod
    def _create_async_cb_wrapper(
        callback: AsyncCallable, *args: Any, **kwds: Any
    ) -> AsyncCallable:
        async def _exit_wrapper(
            exc_type: Type, exc: BaseException, tb: types.TracebackType
        ) -> None:
            await callback(*args, **kwds)

        return _exit_wrapper

    async def enter_async_context(self, cm: AsyncContextManager) -> Any:
        """Enters the supplied async context manager.

        If successful, also pushes its __aexit__ method as a callback and
        returns the result of the __aenter__ method.
        """
        _cm_type = type(cm)
        _exit = _cm_type.__aexit__
        result = await _cm_type.__aenter__(cm)
        self._push_async_cm_exit(cm, _exit)
        return result

    def push_async_exit(self, exit: AsyncPushArg) -> AsyncPushArg:
        """Register coroutine with the standard __aexit__ method signature.

        Can suppress exceptions the same way __aexit__ method can.
        Also accepts any object with an __aexit__ method (registering a call
        to the method instead of the object itself).
        """
        _cb_type = type(exit)
        try:
            exit_method = _cb_type.__aexit__  # type: ignore
        except AttributeError:
            # Not an async context manager, so assume it's a coroutine function
            self._push_exit_callback(exit, False)  # type: ignore
        else:
            self._push_async_cm_exit(exit, exit_method)  # type: ignore
        return exit  # Allow use as a decorator

    def push_async_callback(
        self, callback: AsyncCallable, *args: Any, **kwds: Any
    ) -> AsyncCallable:
        """Register an arbitrary coroutine function and arguments.

        Cannot suppress exceptions.
        """
        _exit_wrapper = self._create_async_cb_wrapper(callback, *args, **kwds)

        # We changed the signature, so using @wraps is not appropriate, but
        # setting __wrapped__ may still help with introspection.
        _exit_wrapper.__wrapped__ = callback  # type: ignore
        self._push_exit_callback(_exit_wrapper, False)
        return callback  # Allow use as a decorator

    async def aclose(self) -> None:
        """Immediately unwind the context stack."""
        await self.__aexit__(None, None, None)

    def _push_async_cm_exit(self, cm: AsyncContextManager, cm_exit: Callable) -> None:
        # Helper to correctly register coroutine function to __aexit__
        # method.
        _exit_wrapper = self._create_async_exit_wrapper(cm, cm_exit)
        _exit_wrapper.__self__ = cm  # type: ignore
        self._push_exit_callback(_exit_wrapper, False)

    async def __aenter__(self) -> "AsyncExitStack":
        return self

    async def __aexit__(self, *exc_details: Any) -> Any:
        received_exc = exc_details[0] is not None

        # We manipulate the exception state so it behaves as though
        # we were actually nesting multiple with statements
        frame_exc = sys.exc_info()[1]

        def _fix_exception_context(
            new_exc: BaseException, old_exc: BaseException
        ) -> None:
            # Context may not be correct, so find the end of the chain
            while 1:
                exc_context = new_exc.__context__
                if exc_context is old_exc:
                    # Context is already set correctly (see issue 20317)
                    return
                if exc_context is None or exc_context is frame_exc:
                    break
                new_exc = exc_context
            # Change the end of the chain to point to the exception
            # we expect it to reference
            new_exc.__context__ = old_exc

        # Callbacks are invoked in LIFO order to match the behaviour of
        # nested context managers
        suppressed_exc = False
        pending_raise = False
        while self._exit_callbacks:
            is_sync, cb = self._exit_callbacks.pop()
            try:
                if is_sync:
                    cb_suppress = cb(*exc_details)
                else:
                    cb_suppress = await cb(*exc_details)

                if cb_suppress:
                    suppressed_exc = True
                    pending_raise = False
                    exc_details = (None, None, None)
            except:  # noqa
                new_exc_details = sys.exc_info()
                # simulate the stack of exceptions by setting the context
                _fix_exception_context(
                    new_exc_details[1], exc_details[1]  # type: ignore
                )
                pending_raise = True
                exc_details = new_exc_details
        if pending_raise:
            try:
                # bare "raise exc_details[1]" replaces our carefully
                # set-up context
                fixed_ctx = exc_details[1].__context__
                raise exc_details[1]
            except BaseException:
                exc_details[1].__context__ = fixed_ctx
                raise
        return received_exc and suppressed_exc


class nullcontext(AbstractContextManager):
    """Context manager that does no additional processing.

    Used as a stand-in for a normal context manager, when a particular
    block of code is only sometimes used with a normal context manager:

    cm = optional_cm if condition else nullcontext()
    with cm:
        # Perform operation, using optional_cm if condition is True
    """

    def __init__(self, enter_result: Any = None) -> None:
        self.enter_result: Any = enter_result

    def __enter__(self) -> Any:
        return self.enter_result

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[types.TracebackType],
    ) -> None:
        pass



import pytest

from mode.utils.contexts import asyncnullcontext


@pytest.mark.asyncio
async def test_asyncnullcontext():
    ctx = asyncnullcontext(enter_result=30)
    async with ctx as value:
        assert value == 30



"""ServiceThread - Service that starts in a separate thread.

Will use the default thread pool executor (``loop.set_default_executor()``),
unless you specify a specific executor instance.

Note: To stop something using the thread's loop, you have to
use the ``on_thread_stop`` callback instead of the on_stop callback.
"""
import asyncio
import sys
import threading
import traceback
from asyncio.locks import Event
from time import monotonic
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Type,
)

from .services import Service
from .utils.futures import maybe_async, maybe_set_exception, maybe_set_result, notify

__all__ = [
    "QueuedMethod",
    "WorkerThread",
    "ServiceThread",
    "QueueServiceThread",
]


class QueuedMethod(NamedTuple):
    """Describe a method to be called by thread."""

    promise: asyncio.Future
    method: Callable[..., Awaitable[Any]]
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]


class WorkerThread(threading.Thread):
    """Thread class used for services running in a dedicated thread."""

    service: "ServiceThread"
    is_stopped: threading.Event

    def __init__(self, service: "ServiceThread", **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.service = service
        self.daemon = False
        self.is_stopped = threading.Event()

    def run(self) -> None:
        try:
            self.service._start_thread()
        finally:
            self._set_stopped()

    def _set_stopped(self) -> None:
        try:
            self.is_stopped.set()
        except TypeError:  # pragma: no cover
            # we lost the race at interpreter shutdown,
            # so gc collected built-in modules.
            pass

    def stop(self) -> None:
        self.is_stopped.wait()
        if self.is_alive():
            self.join(threading.TIMEOUT_MAX)


class ServiceThread(Service):
    """Service subclass running within a dedicated thread."""

    Worker: Type[WorkerThread] = WorkerThread

    abstract = True
    wait_for_shutdown = True

    #: Set this to False if s.start() should not wait for the
    #: underlying thread to be fully started.
    wait_for_thread: bool = True

    _thread: Optional["WorkerThread"] = None
    _thread_started: Event
    _thread_running: Optional[asyncio.Future] = None

    last_wakeup_at: float = 0.0

    def __init__(
        self,
        *,
        executor: Any = None,
        loop: asyncio.AbstractEventLoop = None,
        thread_loop: asyncio.AbstractEventLoop = None,
        Worker: Type[WorkerThread] = None,
        **kwargs: Any,
    ) -> None:
        # cannot share loop between threads, so create a new one
        if executor is not None:
            raise NotImplementedError("executor argument no longer supported")
        self.parent_loop = loop or asyncio.get_event_loop_policy().get_event_loop()
        self.thread_loop = (
            thread_loop or asyncio.get_event_loop_policy().new_event_loop()
        )
        self._thread_started = Event()
        if Worker is not None:
            self.Worker = Worker
        super().__init__(loop=self.thread_loop, **kwargs)

    async def on_thread_started(self) -> None:
        ...

    async def on_thread_stop(self) -> None:
        ...

    # The deal with asyncio.Event and threads.
    #
    # Every thread needs a dedicated event loop, but events can actually
    # be shared between threads in some ways:
    #
    #   - Any thread can set/check the flag (.set() / .is_set())
    #   - Only the thread owning the loop can wait for the event
    #     to be set (await .wait())

    # So X(Service) adds dependency Y(ServiceThread)

    # We add a new _thread_started event owned by the parent loop.
    #
    # Original ._started event is owned by parent loop
    #
    #    X calls await Y.start(): this starts a thread running Y._start_thread
    #    Y starts the thread, and the thread calls super().start to start
    #    the ServiceThread inside that thread.
    #    After starting the thread will wait for _stopped to be set.

    # ._stopped is owned by thread loop
    #      parent sets _stopped.set(), thread calls _stopped.wait()
    #      and only wait needs the loop.
    # ._shutdown is owned by parent loop
    #      thread calls _shutdown.set(), parent calls _shutdown.wait()

    def _new_shutdown_event(self) -> Event:
        return Event()

    async def maybe_start(self) -> bool:
        if not self._thread_started.is_set():
            await self.start()
            return True
        return False

    async def start(self) -> None:
        assert not self._thread_started.is_set()
        self._thread_started.set()
        self._thread_running = asyncio.Future(loop=self.parent_loop)
        self.add_future(self._keepalive2())
        try:
            self._thread = self.Worker(self)
            self._thread.start()
            if not self.should_stop and self.wait_for_thread:
                # thread exceptions do not propagate to the main thread,
                # so we need some way to communicate socket open errors,
                # such as "port in use", back to the parent thread.
                # The _thread_running future is set to
                # an exception state when that happens, and awaiting will
                # propagate the error to the parent thread.

                # wait for thread to be fully started
                await self._thread_running
        finally:
            self._thread_running = None

    async def _keepalive2(self) -> None:
        while not self.should_stop:
            await self.sleep(1.1)
            if self.last_wakeup_at:
                if monotonic() - self.last_wakeup_at > 3.0:
                    self.log.error("Thread keepalive is not responding...")
            asyncio.run_coroutine_threadsafe(
                self._wakeup_timer_in_thread(), self.thread_loop
            )

    async def _wakeup_timer_in_thread(self) -> None:
        self.last_wakeup_at = monotonic()
        await self.sleep(0)
        asyncio.run_coroutine_threadsafe(asyncio.sleep(0), self.parent_loop)

    async def crash(self, exc: BaseException) -> None:
        # <- .start() will raise
        if asyncio.get_event_loop_policy().get_event_loop() is self.parent_loop:
            maybe_set_exception(self._thread_running, exc)
        else:
            self.parent_loop.call_soon_threadsafe(
                maybe_set_exception, self._thread_running, exc
            )
        await super().crash(exc)

    def _start_thread(self) -> None:
        # set the default event loop for this thread
        asyncio.set_event_loop(self.thread_loop)
        try:
            self.thread_loop.run_until_complete(self._serve())
        except Exception:
            # if self._serve raises an exception we need to set
            # shutdown here, since _shutdown_thread will not execute.
            self.set_shutdown()
            raise

    async def stop(self) -> None:
        if self._started.is_set():
            await super().stop()
            if self._thread is not None:
                self._thread.stop()

    def _stopped_set(self) -> None:
        self.thread_loop.call_soon_threadsafe(self._stopped.set)

    def set_shutdown(self) -> None:
        self.parent_loop.call_soon_threadsafe(self._shutdown.set)

    async def _stop_children(self) -> None:
        ...  # called by thread instead of .stop()

    async def _stop_futures(self) -> None:
        ...  # called by thread instead of .stop()

    async def _stop_exit_stacks(self) -> None:
        ...  # called by thread instead of .stop()

    async def _shutdown_thread(self) -> None:
        await self.on_thread_stop()
        await self._default_stop_children()
        self.set_shutdown()
        await self._default_stop_futures()
        await self._default_stop_exit_stacks()

    async def _serve(self) -> None:
        try:
            # start the service
            await self._default_start()
            # allow ServiceThread.start() to return
            # when wait_for_thread is enabled.
            await self.on_thread_started()
            self.parent_loop.call_soon_threadsafe(notify, self._thread_running)
            await self.wait_until_stopped()
        except asyncio.CancelledError:
            raise
        except BaseException as exc:  # pylint: disable=broad-except
            self.on_crash("{0!r} crashed: {1!r}", self.label, exc)
            await self.crash(exc)
            if self.beacon.root is not None:
                await self.beacon.root.data.crash(exc)
            raise
        finally:
            await self._shutdown_thread()

    @Service.task
    async def _thread_keepalive(self) -> None:
        async for sleep_time in self.itertimer(
            1.0, name=f"_thread_keepalive-{self.label}", loop=self.thread_loop
        ):  # pragma: no cover
            # The consumer thread will have a separate event loop,
            # and so we use this trick to make sure our loop is
            # being scheduled to run something at all times.
            #
            # If we don't do this, anything waiting for new
            # stuff in the method queue may never get it.
            pass

    def on_crash(self, msg: str, *fmt: Any, **kwargs: Any) -> None:
        print(msg.format(*fmt), file=sys.stderr)
        traceback.print_exc(None, sys.stderr)


class MethodQueueWorker(Service):
    index: int
    method_queue: "MethodQueue"
    mundane_level = "debug"

    def __init__(
        self, method_queue: "MethodQueue", *, index: int, **kwargs: Any
    ) -> None:
        self.method_queue = method_queue
        self.index = index
        super().__init__(**kwargs)

    @Service.task
    async def _method_queue_do_work(self) -> None:
        method_queue = self.method_queue
        queue_ready = method_queue._queue_ready
        wait = method_queue.wait
        get = method_queue._queue.get
        process_enqueued = method_queue._process_enqueued
        while not self.should_stop and not method_queue.should_stop:
            await wait(queue_ready)

            if not method_queue.should_stop:
                item = await get()
                await process_enqueued(item)

    @property
    def label(self) -> str:
        return f"{type(self).__name__}@{id(self):#x} index={self.index}"


class MethodQueue(Service):
    Worker: Type[MethodQueueWorker] = MethodQueueWorker

    _queue: asyncio.Queue
    _queue_ready: Event
    _workers: List[MethodQueueWorker]

    mundane_level = "debug"

    def __init__(self, num_workers: int = 2, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._queue = asyncio.Queue()
        self._queue_ready = Event()
        self.num_workers = num_workers
        self._workers = []

    async def on_start(self) -> None:
        self._workers[:] = [
            self.Worker(self, index=i, loop=self.loop, beacon=self.beacon)
            for i in range(self.num_workers)
        ]
        for worker in self._workers:
            await self.add_runtime_dependency(worker)

    async def on_stop(self) -> None:
        await self.flush()
        self._workers[:] = []

    async def call(
        self,
        promise: asyncio.Future,
        fun: Callable[..., Awaitable],
        *args: Any,
        **kwargs: Any,
    ) -> asyncio.Future:
        method = QueuedMethod(promise, fun, args, kwargs)
        self.loop.call_soon_threadsafe(self._queue_put, method)
        return promise

    def _queue_put(self, method: QueuedMethod) -> None:
        self._queue.put_nowait(method)
        self._queue_ready.set()

    async def cast(
        self, fun: Callable[..., Awaitable], *args: Any, **kwargs: Any
    ) -> None:
        promise = self.loop.create_future()
        method = QueuedMethod(promise, fun, args, kwargs)
        self._queue.put_nowait(method)
        self._queue_ready.set()

    async def flush(self) -> None:
        while 1:
            try:
                p = self._queue.get_nowait()
            except asyncio.QueueEmpty:
                break
            else:
                await self._process_enqueued(p)

    async def _process_enqueued(self, p: QueuedMethod) -> asyncio.Future:
        promise, method, args, kwargs = p
        if not promise.cancelled():
            try:
                result = await maybe_async(method(*args, **kwargs))
            except BaseException as exc:
                promise._loop.call_soon_threadsafe(maybe_set_exception, promise, exc)
            else:
                promise._loop.call_soon_threadsafe(maybe_set_result, promise, result)
        return promise

    @property
    def label(self) -> str:
        return f"{type(self).__name__}@{id(self):#x}"


class QueueServiceThread(ServiceThread):
    """Service running in separate thread.

    Uses a queue to run functions inside the thread,
    so you can delegate calls.
    """

    abstract = True

    _method_queue: Optional[MethodQueue] = None

    @property
    def method_queue(self) -> MethodQueue:
        if self._method_queue is None:
            self._method_queue = MethodQueue(
                loop=self.thread_loop,
                beacon=self.beacon,
            )
        return self._method_queue

    async def on_thread_started(self) -> None:
        await self.method_queue.start()

    async def on_thread_stop(self) -> None:
        if self._method_queue is not None:
            await self._method_queue.stop()

    async def call_thread(
        self, fun: Callable[..., Awaitable], *args: Any, **kwargs: Any
    ) -> Any:
        # Enqueue method to be called by thread (synchronous).

        # We pass a future to the thread, so that when the call is done
        # the thread will call `future.set_result(result)`.
        promise = await self.method_queue.call(
            self.parent_loop.create_future(), fun, *args, **kwargs
        )

        # wait for the promise to be fulfilled
        result = await promise
        return result

    async def cast_thread(
        self, fun: Callable[..., Awaitable], *args: Any, **kwargs: Any
    ) -> None:
        # Enqueue method to be called by thread (asynchronous).
        await self.method_queue.cast(fun, *args, **kwargs)



"""Async I/O services that can be started/stopped/shutdown."""
import asyncio
import logging
import sys
from asyncio.locks import Event
from contextlib import AsyncExitStack, ExitStack
from datetime import tzinfo
from functools import wraps
from time import monotonic, perf_counter
from types import TracebackType
from typing import (
    Any,
    AsyncContextManager,
    AsyncIterator,
    Awaitable,
    Callable,
    ClassVar,
    ContextManager,
    Coroutine,
    Dict,
    Iterable,
    List,
    Mapping,
    MutableSequence,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Type,
    Union,
    cast,
)

from .timers import Timer
from .types import DiagT, ServiceT
from .utils.cron import secs_for_next
from .utils.logging import CompositeLogger, get_logger, level_number
from .utils.objects import iter_mro_reversed, qualname
from .utils.text import maybecat
from .utils.times import Seconds, want_seconds
from .utils.tracebacks import format_task_stack
from .utils.trees import Node
from .utils.types.trees import NodeT

__all__ = ["ServiceBase", "Service", "Diag", "task", "timer", "crontab"]

ClockArg = Callable[[], float]

#: Future type: Different types of awaitables.
FutureT = Union[asyncio.Future, Coroutine[Any, None, Any], Awaitable]

#: Argument type for ``Service.wait(*events)``
#: Wait can take any number of futures or events to wait for.
WaitArgT = Union[FutureT, Event]


class WaitResults(NamedTuple):
    done: List[WaitArgT]
    results: List[Any]
    stopped: bool


class WaitResult(NamedTuple):
    """Return value of :meth:`Service.wait`."""

    #: Return value of the future we were waiting for.
    result: Any

    #: Set to :const:`True` if the service was stopped while waiting.
    stopped: bool


class ServiceBase(ServiceT):
    """Base class for services."""

    # This class implements stuff common to Service + ServiceProxy

    #: Set to True if this service class is abstract-only,
    #: meaning it will only be used as a base class.
    abstract: ClassVar[bool] = True

    log: CompositeLogger

    #: Logger used by this service.
    #: If not explicitly set this will be based on get_logger(cls.__name__)
    # This is automatically set when class is constructed, and so is only
    # None on the class, never on an instance. For simplicity we cast
    # the None to logger.
    logger: logging.Logger = cast(logging.Logger, None)

    def __init_subclass__(cls) -> None:
        if cls.abstract:
            cls.abstract = False
        cls._init_subclass_logger()

    @classmethod
    def _init_subclass_logger(cls) -> None:
        # make sure class has a logger.
        logger = cast(Optional[logging.Logger], cls.logger)
        if logger is None or getattr(logger, "__modex__", False):
            _logger = cls.logger = get_logger(cls.__module__)
            _logger.__modex__ = True  # type: ignore

    def __init__(self, *, loop: asyncio.AbstractEventLoop = None) -> None:
        self.log = CompositeLogger(self.logger, formatter=self._format_log)
        self._loop = loop

    def _format_log(self, severity: int, msg: str, *args: Any, **kwargs: Any) -> str:
        return f'[^{"-" * (self.beacon.depth - 1)}{self.shortlabel}]: {msg}'

    async def __aenter__(self) -> ServiceT:
        await self.start()
        return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException] = None,
        exc_val: BaseException = None,
        exc_tb: TracebackType = None,
    ) -> Optional[bool]:
        try:
            await self.stop()
        finally:
            self.service_reset()
        return None

    def __repr__(self) -> str:
        # Override _repr_info to add additional text to repr.
        info = maybecat(self._repr_info(), prefix=" ") or ""
        return f"<{self._repr_name()}: {self.state}{info}>"

    def _repr_info(self) -> str:
        return ""

    def _repr_name(self) -> str:
        return type(self).__name__

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if self._loop is None:
            self._loop = asyncio.get_event_loop_policy().get_event_loop()
        return self._loop

    @loop.setter
    def loop(self, loop: asyncio.AbstractEventLoop) -> None:
        self._loop = loop


class Diag(DiagT):
    """Service diagnostics.

    This can be used to track what your service is doing.
    For example if your service is a Kafka consumer with a background
    thread that commits the offset every 30 seconds, you may want to
    see when this happens::

        DIAG_COMMITTING = 'committing'

        class Consumer(Service):

            @Service.task
            async def _background_commit(self) -> None:
                while not self.should_stop:
                    await self.sleep(30.0)
                    self.diag.set_flag(DIAG_COMITTING)
                    try:
                        await self._consumer.commit()
                    finally:
                        self.diag.unset_flag(DIAG_COMMITTING)

    The above code is setting the flag manually, but you can also use
    a decorator to accomplish the same thing::

        @Service.timer(30.0)
        async def _background_commit(self) -> None:
            await self.commit()

        @Service.transitions_with(DIAG_COMITTING)
        async def commit(self) -> None:
            await self._consumer.commit()
    """

    def __init__(self, service: ServiceT) -> None:
        self.service = service
        self.flags = set()
        self.last_transition = {}

    def set_flag(self, flag: str) -> None:
        self.flags.add(flag)
        self.last_transition[flag] = monotonic()

    def unset_flag(self, flag: str) -> None:
        self.flags.discard(flag)


class ServiceTask:
    """A background task.

    You don't have to use this class directly, instead
    use the ``@Service.task`` decorator::

        class MyService(Service):

            @Service.task
            def _background_task(self):
                while not self.should_stop:
                    print('Hello')
                    await self.sleep(1.0)
    """

    def __init__(self, fun: Callable[..., Awaitable]) -> None:
        self.fun: Callable[..., Awaitable] = fun

    async def __call__(self, obj: Any) -> Any:
        return await self.fun(obj)

    def __repr__(self) -> str:
        return repr(self.fun)


class ServiceCallbacks:
    """Service callback interface.

    When calling ``await service.start()`` this happens:

    .. code-block:: text

        +--------------------+
        | INIT (not started) |
        +--------------------+
                V
        .-----------------------.
        / await service.start() |
        `-----------------------'
                V
        +--------------------+
        | on_first_start     |
        +--------------------+
                V
        +--------------------+
        | on_start           |
        +--------------------+
                V
        +--------------------+
        | on_started         |
        +--------------------+

    When stopping and ``wait_for_shutdown`` is unset, this happens:

    .. code-block:: text

        .-----------------------.
        / await service.stop()  |
        `-----------------------'
                V
        +--------------------+
        | on_stop             |
        +--------------------+
                V
        +--------------------+
        | on_shutdown        |
        +--------------------+

    When stopping and ``wait_for_shutdown`` is set, the stop operation
    will wait for something to set the shutdown flag ``self.set_shutdown()``:

    .. code-block:: text

        .-----------------------.
        / await service.stop()  |
        `-----------------------'
                V
        +--------------------+
        | on_stop             |
        +--------------------+
                V
        .-------------------------.
        / service.set_shutdown()  |
        `-------------------------'
                V
        +--------------------+
        | on_shutdown        |
        +--------------------+

    When restarting the order is as follows (assuming
    ``wait_for_shutdown`` unset):

    .. code-block:: text

        .-------------------------.
        / await service.restart() |
        `-------------------------'
                V
        +--------------------+
        | on_stop             |
        +--------------------+
                V
        +--------------------+
        | on_shutdown        |
        +--------------------+
                V
        +--------------------+
        | on_restart         |
        +--------------------+
                V
        +--------------------+
        | on_start           |
        +--------------------+
                V
        +--------------------+
        | on_started         |
        +--------------------+
    """

    async def on_first_start(self) -> None:
        """Service started for the first time in this process."""
        ...

    async def on_start(self) -> None:
        """Service is starting."""
        ...

    async def on_started(self) -> None:
        """Service has started."""
        ...

    async def on_stop(self) -> None:
        """Service is being stopped/restarted."""
        ...

    async def on_shutdown(self) -> None:
        """Service is being stopped/restarted."""
        ...

    async def on_restart(self) -> None:
        """Service is being restarted."""
        ...


class Service(ServiceBase, ServiceCallbacks):
    """An asyncio service that can be started/stopped/restarted.

    Keyword Arguments:
        beacon (NodeT): Beacon used to track services in a graph.
        loop (asyncio.AbstractEventLoop): Event loop object.
    """

    abstract: ClassVar[bool] = True
    Diag: Type[DiagT] = Diag

    #: Set to True if .stop must wait for the shutdown flag to be set.
    wait_for_shutdown = False

    #: Time to wait for shutdown flag set before we give up.
    shutdown_timeout = 60.0

    #: Current number of times this service instance has been restarted.
    restart_count = 0

    #: The log level for mundane info such as `starting`, `stopping`, etc.
    #: Set this to ``"debug"`` for less information.
    mundane_level = "info"
    _mundane_level: int

    #: Event set when service started.
    _started: Event

    #: Event set when service stopped.
    _stopped: Event

    #: Event set by user to signal service can be shutdown
    #: (see :attr:`wait_for_shutdown)
    _shutdown: Event

    #: Event set when service crashed.
    _crashed: Event

    #: The reason for last crash (an exception instance).
    _crash_reason: Optional[BaseException]

    #: The beacon is used to maintain a graph of services.
    _beacon: NodeT

    #: .add_dependency and friends adds services to this list,
    #: that are started/stopped/restarted with the service.
    _children: MutableSequence[ServiceT]

    #: .add_future adds futures to this list, and when stopping
    #: we will wait for them a bit, then cancel them.
    #: Note: Unlike ``add_dependency`` these futures will not be
    # restarted with the service: if you want that to happen make sure
    # calling service.start() again will add the future again.
    _futures: Set[asyncio.Future]

    #: The ``@Service.task`` decorator adds names of attributes
    #: that are ServiceTasks to this list (which is a class variable).
    _tasks: ClassVar[Optional[Dict[str, Set[str]]]] = None

    @classmethod
    def from_awaitable(
        cls, coro: Awaitable, *, name: str = None, **kwargs: Any
    ) -> ServiceT:
        return _AwaitableService(coro, name=name)

    @classmethod
    def task(cls, fun: Callable[[Any], Awaitable[None]]) -> ServiceTask:
        """Decorate function to be used as background task.

        Example:
            >>> class S(Service):
            ...
            ...     @Service.task
            ...     async def background_task(self):
            ...         while not self.should_stop:
            ...             await self.sleep(1.0)
            ...             print('Waking up')
        """
        return ServiceTask(fun)

    @classmethod
    def timer(
        cls,
        interval: Seconds,
        *,
        exec_first: bool = False,
        name: str = None,
        max_drift_correction: float = 0.1,
    ) -> Callable[[Callable], ServiceTask]:
        """Background timer executing every ``n`` seconds.

        Example:
            >>> class S(Service):
            ...
            ...     @Service.timer(1.0)
            ...     async def background_timer(self):
            ...         print('Waking up')
        """
        _interval = want_seconds(interval)

        def _decorate(fun: Callable[[ServiceT], Awaitable[None]]) -> ServiceTask:
            _timer_name = name or qualname(fun)

            @wraps(fun)
            async def _repeater(self: Service) -> None:
                if exec_first:
                    await fun(self)
                async for sleep_time in self.itertimer(
                    _interval,
                    name=_timer_name,
                    max_drift_correction=max_drift_correction,
                ):
                    await fun(self)

            return cls.task(_repeater)

        return _decorate

    @classmethod
    def crontab(
        cls, cron_format: str, *, timezone: tzinfo = None
    ) -> Callable[[Callable], ServiceTask]:
        """Background timer executing periodic task based on Crontab description.

        Example:
            >>> class S(Service):
            ...
            ...     @Service.crontab(cron_format='30 18 * * *',
                                     timezone=pytz.timezone('US/Pacific'))
            ...     async def every_6_30_pm_pacific(self):
            ...         print('IT IS 6:30pm')
            ...
            ...     @Service.crontab(cron_format='30 18 * * *')
            ...     async def every_6_30_pm(self):
            ...         print('6:30pm UTC')
        """

        def _decorate(fun: Callable[[ServiceT], Awaitable[None]]) -> ServiceTask:
            @wraps(fun)
            async def _cron_starter(self: Service) -> None:
                while not self.should_stop:
                    await self.sleep(secs_for_next(cron_format, timezone))
                    if not self.should_stop:
                        await fun(self)

            return cls.task(_cron_starter)

        return _decorate

    @classmethod
    def transitions_to(cls, flag: str) -> Callable:
        """Decorate function to set and reset diagnostic flag."""

        def _decorate(fun: Callable[..., Awaitable]) -> Callable[..., Awaitable]:
            @wraps(fun)
            async def _and_transition(self: ServiceT, *args: Any, **kwargs: Any) -> Any:
                self.diag.set_flag(flag)
                try:
                    return await fun(self, *args, **kwargs)
                finally:
                    self.diag.unset_flag(flag)

            return _and_transition

        return _decorate

    def __init_subclass__(cls) -> None:
        # Every new subclass adds @Service.task decorated methods
        # to the class-local `_tasks` list.
        if cls.abstract:
            cls.abstract = False
        cls._init_subclass_logger()
        cls._init_subclass_tasks()

    @classmethod
    def _init_subclass_tasks(cls) -> None:
        # XXX Python 3.6.3 introduces mysterious bug
        # where the storage for subclasses is always the same,
        # so when we set cls._tasks = [], it will actually clear the
        # tasks for all subclasses.  Hacked around this situation
        # by making _tasks a mapping from classid -> list of tasks,
        # that way all subclasses can share the same attribute.
        clsid = cls._get_class_id()
        if cls._tasks is None:
            cls._tasks = {}
        tasks: Set[str] = set()
        for base in iter_mro_reversed(cls, stop=Service):
            tasks |= {
                attr_name
                for attr_name, attr_value in vars(base).items()
                if isinstance(attr_value, ServiceTask)
            }
        cls._tasks[clsid] = tasks

    def _get_tasks(self) -> Iterable[ServiceTask]:
        seen: Set[ServiceTask] = set()
        cls = type(self)
        if cls._tasks:
            for attr_name in cls._tasks[cls._get_class_id()]:
                task = getattr(self, attr_name)
                assert isinstance(task, ServiceTask)
                assert task not in seen
                seen.add(task)
                yield task

    @classmethod
    def _get_class_id(cls) -> str:
        return ".".join([cls.__module__, cls.__qualname__])

    def __init__(
        self, *, beacon: NodeT = None, loop: asyncio.AbstractEventLoop = None
    ) -> None:
        self.diag = self.Diag(self)
        self._loop = loop
        self._started = self._new_started_event()
        self._stopped = self._new_stopped_event()
        self._shutdown = self._new_shutdown_event()
        self._crashed = self._new_crashed_event()
        self._crash_reason = None
        self._beacon = Node(self) if beacon is None else beacon.new(self)
        self._children = []
        self._futures = set()
        self._mundane_level = level_number(self.mundane_level)
        self.async_exit_stack = AsyncExitStack()
        self.exit_stack = ExitStack()
        self.on_init()
        self.__post_init__()
        super().__init__(loop=self._loop)

    def _new_started_event(self) -> Event:
        return Event()

    def _new_stopped_event(self) -> Event:
        return Event()

    def _new_shutdown_event(self) -> Event:
        return Event()

    def _new_crashed_event(self) -> Event:
        return Event()

    async def transition_with(
        self, flag: str, fut: Awaitable, *args: Any, **kwargs: Any
    ) -> Any:
        self.diag.set_flag(flag)
        try:
            return await fut
        finally:
            self.diag.unset_flag(flag)

    def add_dependency(self, service: ServiceT) -> ServiceT:
        """Add dependency to other service.

        The service will be started/stopped with this service.
        """
        if service.beacon is not None:
            service.beacon.reattach(self.beacon)
        self._children.append(service)
        return service

    async def add_runtime_dependency(self, service: ServiceT) -> ServiceT:
        self.add_dependency(service)
        if self._started.is_set():
            await service.maybe_start()
        return service

    async def remove_dependency(self, service: ServiceT) -> ServiceT:
        """Stop and remove dependency of this service."""
        await service.stop()
        self._children.remove(service)
        if service.beacon is not None:
            service.beacon.detach(self.beacon)
        return service

    async def add_async_context(self, context: AsyncContextManager) -> Any:
        if isinstance(context, AsyncContextManager):
            return await self.async_exit_stack.enter_async_context(context)
        elif isinstance(context, ContextManager):  # type: ignore
            raise TypeError("Use `self.add_context(ctx)` for non-async context")
        raise TypeError(f"Not a context/async context: {type(context)!r}")

    def add_context(self, context: ContextManager) -> Any:
        if isinstance(context, AsyncContextManager):
            raise TypeError("Use `await self.add_async_context(ctx)` for async context")
        elif isinstance(context, ContextManager):
            return self.exit_stack.enter_context(context)
        raise TypeError(f"Not a context/async context: {type(context)!r}")

    def add_future(self, coro: Awaitable) -> asyncio.Future:
        """Add relationship to asyncio.Future.

        The future will be joined when this service is stopped.
        """
        fut = asyncio.ensure_future(self._execute_task(coro), loop=self.loop)
        try:
            fut.set_name(repr(coro))
        except AttributeError:
            pass
        fut.__wrapped__ = coro  # type: ignore
        fut.add_done_callback(self._on_future_done)
        self._futures.add(fut)
        return fut

    def _get_task_name(self, t: Any) -> str:
        try:
            return cast(str, t.get_name())
        except AttributeError:
            return repr(t)

    def tracebacks(self) -> Mapping[str, str]:
        return {
            self._get_task_name(f): format_task_stack(cast(asyncio.Task, f))
            for f in self._futures
            if not f.done()
        }

    def human_tracebacks(self) -> str:
        return "\n".join(
            [
                "\n".join(
                    [
                        name,
                        "=" * len(name),
                        tb,
                    ]
                )
                for name, tb in self.tracebacks().items()
            ]
        )

    def _on_future_done(self, fut: asyncio.Future) -> None:
        self._futures.discard(fut)

    def __post_init__(self) -> None:
        """Additional user initialization."""
        ...

    def on_init(self) -> None:
        ...  # deprecated: use __post_init__

    def on_init_dependencies(self) -> Iterable[ServiceT]:
        """Return list of service dependencies for this service."""
        return []

    async def join_services(self, services: Sequence[ServiceT]) -> None:
        for service in services:
            try:
                await service.maybe_start()
            except BaseException as exc:
                await self.crash(exc)
        for service in reversed(services):
            await service.stop()

    async def sleep(self, n: Seconds) -> None:
        """Sleep for ``n`` seconds, or until service stopped."""
        try:
            await asyncio.wait_for(
                self._stopped.wait(),
                timeout=want_seconds(n),
            )
        except asyncio.TimeoutError:
            pass

    async def wait_for_stopped(self, *coros: WaitArgT, timeout: Seconds = None) -> bool:
        return (await self.wait(*coros, timeout=timeout)).stopped

    async def wait(self, *coros: WaitArgT, timeout: Seconds = None) -> WaitResult:
        """Wait for coroutines to complete, or until the service stops."""
        timeout = want_seconds(timeout) if timeout is not None else None
        if coros:
            assert len(coros) == 1
            return await self._wait_one(coros[0], timeout=timeout)
        else:
            await self._wait_stopped(timeout=timeout)
            return WaitResult(None, True)

    async def wait_many(
        self, coros: Iterable[WaitArgT], *, timeout: Seconds = None
    ) -> WaitResult:
        timeout = want_seconds(timeout) if timeout is not None else None
        coro = asyncio.wait(
            [
                asyncio.ensure_future(
                    c.wait() if isinstance(c, Event) else c, loop=self.loop
                )
                for c in coros
            ],
            return_when=asyncio.ALL_COMPLETED,
            timeout=timeout,
        )
        return await self._wait_one(coro, timeout=timeout)

    async def wait_first(
        self, *coros: WaitArgT, timeout: Seconds = None
    ) -> WaitResults:
        timeout = want_seconds(timeout) if timeout is not None else None
        stopped = self._stopped
        crashed = self._crashed
        loop = self.loop

        futures = {
            coro: asyncio.ensure_future(
                (coro.wait() if isinstance(coro, Event) else coro),
                loop=loop,
            )
            for coro in coros
        }
        futures[stopped] = asyncio.ensure_future(stopped.wait(), loop=loop)
        futures[crashed] = asyncio.ensure_future(crashed.wait(), loop=loop)
        done: Set[asyncio.Future]
        pending: Set[asyncio.Future]
        try:
            done, pending = await asyncio.wait(
                futures.values(),
                return_when=asyncio.FIRST_COMPLETED,
                timeout=timeout,
            )
            for f in done:
                if f.done() and f.exception() is not None:
                    f.result()  # propagate exceptions
            winners: List[WaitArgT] = []
            results: List[Any] = []
            for coro, fut in futures.items():
                if fut.done():
                    winners.append(coro)
                    results.append(fut.result())
                elif fut.cancelled():
                    raise asyncio.CancelledError()
            if winners and not stopped.is_set() and not crashed.is_set():
                return WaitResults(winners, results, False)
            else:
                return WaitResults([], [], True)
        finally:
            # cleanup
            for fut in futures.values():
                if not fut.done():
                    fut.cancel()

    async def _wait_one(self, coro: WaitArgT, *, timeout: Seconds = None) -> WaitResult:
        results = await self.wait_first(coro, timeout=timeout)
        if results.stopped:
            return WaitResult(None, True)
        return WaitResult(results.results[0], False)

    async def _wait_stopped(self, timeout: Seconds = None) -> None:
        timeout = want_seconds(timeout) if timeout is not None else None
        stopped = asyncio.ensure_future(self._stopped.wait(), loop=self.loop)
        crashed = asyncio.ensure_future(self._crashed.wait(), loop=self.loop)
        done, pending = await asyncio.wait(
            [stopped, crashed],
            return_when=asyncio.FIRST_COMPLETED,
            timeout=timeout,
        )
        for fut in done:
            fut.result()  # propagate exceptions
        for fut in pending:
            fut.cancel()
        assert self._crashed.is_set() or self._stopped.is_set()

    async def start(self) -> None:
        await self._default_start()

    async def _default_start(self) -> None:
        loop = self.loop
        assert loop  # make sure loop is set
        assert not self._started.is_set()
        self._started.set()
        await self._actually_start()

    async def _actually_start(self) -> None:
        """Start the service."""
        for _ in [1]:  # to use break
            if not self.restart_count:
                for dep in self.on_init_dependencies():
                    self.add_dependency(dep)
                await self.on_first_start()
                if self.should_stop:
                    break
            self.exit_stack.__enter__()
            await self.async_exit_stack.__aenter__()
            if self.should_stop:
                break
            try:
                self._log_mundane("Starting...")
                await self.on_start()
                if self.should_stop:
                    break
                for task in self._get_tasks():
                    self.add_future(task.fun(self))
                for child in self._children:
                    if child is not None:
                        await child.maybe_start()
                    if self.should_stop:
                        break
                self.log.debug("Started.")
                await self.on_started()
            except BaseException:
                self.exit_stack.__exit__(*sys.exc_info())
                await self.async_exit_stack.__aexit__(*sys.exc_info())
                raise

    async def _execute_task(self, task: Awaitable) -> None:
        try:
            await task
        except asyncio.CancelledError:
            if not self.should_stop:
                self._log_mundane("Terminating cancelled task: %r", task)
        except RuntimeError as exc:
            if "Event loop is closed" in str(exc):
                self.log.info("Cancelled task %r: %s", task, exc)
            else:
                await self.crash(exc)
        except BaseException as exc:
            # the exception will be re-raised by the main thread.
            await self.crash(exc)

    async def maybe_start(self) -> bool:
        """Start the service, if it has not already been started."""
        if not self._started.is_set():
            await self.start()
            return True
        return False

    def _log_mundane(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.log.log(self._mundane_level, msg, *args, **kwargs)

    async def crash(self, reason: BaseException) -> None:
        """Crash the service and all child services."""
        self.log.exception("Crashed reason=%r", reason)
        if not self._crashed.is_set():
            # We record the stack by raising the exception.

            if self.supervisor:
                self.supervisor.wakeup()
            else:
                # if the service has no supervisor we go ahead
                # and mark parent nodes as crashed as well.
                root = self.beacon.root
                seen: Set[NodeT] = set()
                for node in self.beacon.walk():
                    if node in seen:
                        self.log.warning("Recursive loop in beacon: %r: %r", node, seen)
                        if root is not None and root.data is not self:
                            cast(Service, root.data)._crash(reason)
                        break
                    seen.add(node)
                    for child in [node.data] + node.children:
                        if isinstance(child, Service):
                            child._crash(reason)
            self._crash(reason)

    def _crash(self, reason: BaseException) -> None:
        self._crashed.set()
        self.crash_reason = reason
        for node in self._children:
            node._crash(reason)

    async def stop(self) -> None:
        """Stop the service."""
        if not self._stopped.is_set():
            self._log_mundane("Stopping...")
            self._stopped_set()
            await self.on_stop()
            await self._stop_children()
            self.log.debug("Shutting down...")
            if self.wait_for_shutdown:
                self.log.debug("Waiting for shutdown")
                await asyncio.wait_for(
                    self._shutdown.wait(),
                    self.shutdown_timeout,
                )
                self.log.debug("Shutting down now")
            await self._stop_futures()
            await self._stop_exit_stacks()
            await self.on_shutdown()
            self.log.debug("-Stopped!")

    def _stopped_set(self) -> None:
        self._stopped.set()

    async def _stop_children(self) -> None:
        await self._default_stop_children()

    async def _default_stop_children(self) -> None:
        for child in reversed(self._children):
            if child is not None:
                try:
                    await asyncio.shield(child.stop())
                except asyncio.CancelledError:
                    pass
                except Exception as exc:
                    self.log.exception("Error while stopping child %r: %r", child, exc)

    async def _stop_futures(self) -> None:
        await self._default_stop_futures()

    async def _stop_exit_stacks(self) -> None:
        await self._default_stop_exit_stacks()

    async def _default_stop_exit_stacks(self) -> None:
        self.exit_stack.__exit__(None, None, None)
        await self.async_exit_stack.__aexit__(None, None, None)

    async def _default_stop_futures(self) -> None:
        await self._wait_for_futures(timeout=0)
        for future in self._futures:
            future.cancel()
        await self._gather_futures()

    async def _gather_futures(self, *, timeout: float = None) -> None:
        while self._futures:
            # Gather all futures added via .add_future
            try:
                await self._maybe_wait_for_futures(timeout=timeout)
            except asyncio.CancelledError:
                continue
            else:
                break
        self._futures.clear()

    async def _maybe_wait_for_futures(self, *, timeout: float = None) -> None:
        if self._futures:
            try:
                await asyncio.shield(self._wait_for_futures(timeout=timeout))
            except ValueError:
                if self._futures:
                    raise
                # race condition:
                # _futures non-empty when loop starts,
                # but empty when asyncio.wait receives it.
            except asyncio.CancelledError:
                pass

    async def _wait_for_futures(self, *, timeout: float = None) -> None:
        if self._futures:
            await asyncio.wait(
                self._futures,
                return_when=asyncio.ALL_COMPLETED,
                timeout=timeout,
            )

    async def restart(self) -> None:
        """Restart this service."""
        await self.stop()
        self.service_reset()
        await self.on_restart()
        await self.start()

    def service_reset(self) -> None:
        self.restart_count += 1
        for ev in (self._started, self._stopped, self._shutdown, self._crashed):
            ev.clear()
        self.crash_reason = None
        for child in self._children:
            if child is not None:
                child.service_reset()

    async def wait_until_stopped(self) -> None:
        """Wait until the service is signalled to stop."""
        await self.wait()

    def set_shutdown(self) -> None:
        """Set the shutdown signal.

        Notes:
            If :attr:`wait_for_shutdown` is set, stopping the service
            will wait for this flag to be set.
        """
        self._shutdown.set()

    async def itertimer(
        self,
        interval: Seconds,
        *,
        max_drift_correction: float = 0.1,
        sleep: Callable[..., Awaitable] = None,
        clock: ClockArg = perf_counter,
        name: str = "",
    ) -> AsyncIterator[float]:
        """Sleep ``interval`` seconds for every iteration.

        This is an async iterator that takes advantage
        of :func:`~mode.timers.Timer` to monitor drift and timer
        oerlap.

        Uses ``Service.sleep`` so exits fast when the service is
        stopped.

        Note:
            Will sleep the full `interval` seconds before returning
            from first iteration.

        Examples:
            >>> async for sleep_time in self.itertimer(1.0):
            ...   print('another second passed, just woke up...')
            ...   await perform_some_http_request()
        """
        sleepfun = sleep or self.sleep
        if self.should_stop:
            return
        try:
            async for sleep_time in Timer(
                interval,
                name=name,
                max_drift_correction=max_drift_correction,
                clock=clock,
                sleep=sleepfun,
            ):
                if self.should_stop:
                    break
                yield sleep_time
                if self.should_stop:
                    break
        finally:
            # this is required to collect the async_generator_athrow()
            # tasks left running after the `async for` block ends.
            await asyncio.sleep(0)

    @property
    def started(self) -> bool:
        """Return :const:`True` if the service was started."""
        return bool(self._started.is_set())

    @property
    def crashed(self) -> bool:
        return bool(self._crashed.is_set())

    @property
    def should_stop(self) -> bool:
        """Return :const:`True` if the service must stop."""
        return bool(self._stopped.is_set() or self._crashed.is_set())

    @property
    def state(self) -> str:
        """Service state - as a human readable string."""
        if self._crashed.is_set():
            return "crashed"
        elif not self._started.is_set():
            return "init"
        elif not self._stopped.is_set():
            return "running"
        elif not self._shutdown.is_set():
            return "stopping"
        else:
            return "shutdown"

    @property
    def label(self) -> str:
        """Label used for graphs."""
        return self._repr_name()

    @property
    def shortlabel(self) -> str:
        """Label used for logging."""
        return self.label

    @property
    def beacon(self) -> NodeT:
        """Beacon used to track services in a dependency graph."""
        return self._beacon

    @beacon.setter
    def beacon(self, beacon: NodeT) -> None:
        self._beacon = beacon

    @property
    def crash_reason(self) -> Optional[BaseException]:
        return self._crash_reason

    @crash_reason.setter
    def crash_reason(self, reason: Optional[BaseException]) -> None:
        self._crash_reason = reason


task = Service.task
timer = Service.timer
crontab = Service.crontab


class _AwaitableService(Service):
    mundane_level = "debug"

    _fut: Optional[asyncio.Future]

    def __init__(self, coro: Awaitable, *, name: str = None, **kwargs: Any) -> None:
        self.coro = coro
        self._fut = None
        self.name = name
        super().__init__(**kwargs)

    async def on_start(self) -> None:
        # convert to future, so we can cancel on_stop
        try:
            self._fut = asyncio.ensure_future(self.coro, loop=self.loop)
            await self._fut
        except asyncio.CancelledError:
            if not self.should_stop:
                raise

    async def on_stop(self) -> None:
        fut, self._fut = self._fut, None
        if fut is not None:
            if not fut.done():
                fut.cancel()
            else:
                fut.result()

    def _repr_name(self) -> str:
        return self.name or str(self.coro)



"""Data structure: Trees."""
from contextlib import suppress
from typing import Any, Iterator, List, Optional, TypeVar, Union, cast

from .graphs import DependencyGraph
from .objects import shortlabel
from .types.graphs import DependencyGraphT
from .types.trees import NodeT
from .typing import Deque

__all__ = [
    "Node",
]

T = TypeVar("T")


class Node(NodeT[T]):
    """Tree node.

    Notes:
        Nodes have a link to

            - the ``.root`` node (or None if this is the top-most node)
            - the ``.parent`` node (if this is a child node).
            - a list of children

        A Node may have arbitrary ``.data`` associated with it, and arbitrary
        data may also be stored in ``.children``.

    Arguments:
        data (Any): Data to associate with node.

    Keyword Arguments:
        root (NodeT): Root node.
        parent (NodeT): Parent node.
        children (List[NodeT]): List of child nodes.
    """

    _root: Optional[NodeT[T]] = None
    _parent: Optional[NodeT[T]] = None

    @classmethod
    def _new_node(cls, data: T, **kwargs: Any) -> NodeT[T]:
        return cls(data, **kwargs)

    def __init__(
        self,
        data: T,
        *,
        root: NodeT = None,
        parent: NodeT = None,
        children: List[NodeT[T]] = None,
    ) -> None:
        self.data = data
        if root is not None:
            self.root = root
        if parent is not None:
            self.parent = parent
        self.children = children or []

    def new(self, data: T) -> NodeT:
        """Create new node from this node."""
        node = self._new_node(
            data,
            root=self.root if self.root is not None else self,
            parent=self,
        )
        self.children.append(node)
        return node

    def reattach(self, parent: NodeT[T]) -> NodeT[T]:
        """Attach this node to `parent` node."""
        self.root = parent.root if parent.root is not None else parent
        self.parent = parent
        parent.add_deduplicate(self)
        return self

    def detach(self, parent: NodeT[T]) -> NodeT[T]:
        """Detach this node from `parent` node."""
        if self.parent is not None:
            self.parent.discard(self)
        self._parent = None
        self._root = None
        return self

    def add_deduplicate(self, data: Union[T, NodeT[T]]) -> None:
        if data not in self.children:
            self.children.append(data)

    def add(self, data: Union[T, NodeT[T]]) -> None:
        """Add node as a child node."""
        self.children.append(data)

    def discard(self, data: T) -> None:
        """Remove node so it's no longer a child of this node."""
        # XXX slow
        with suppress(ValueError):
            self.children.remove(data)

    def traverse(self) -> Iterator[NodeT[T]]:
        """Iterate over the tree in BFS order."""
        stack: Deque[NodeT[T]] = Deque([self])
        while stack:
            node = stack.popleft()
            yield node
            for child in node.children:
                if isinstance(child, NodeT):
                    stack.append(child)
                else:
                    yield child

    def walk(self) -> Iterator[NodeT[T]]:
        """Iterate over hierarchy backwards.

        This will yield parent nodes all the way up to the root.
        """
        node: Optional[NodeT[T]] = self
        while node:
            yield node
            node = node.parent

    def as_graph(self) -> DependencyGraphT:
        """Convert to :class:`~mode.utils.graphs.DependencyGraph`."""
        graph = DependencyGraph()
        stack: Deque[NodeT] = Deque([self])
        while stack:
            node = stack.popleft()
            for child in node.children:
                graph.add_arc(node.data)
                if isinstance(child, NodeT):
                    stack.append(cast(Node, child))
                    graph.add_edge(node.data, child.data)
                else:
                    graph.add_edge(node.data, child)
        return graph

    def __repr__(self) -> str:
        return f"{type(self).__name__}: {self.path}"

    @property
    def depth(self) -> int:
        return self._find_depth()

    def _find_depth(self) -> int:
        return sum(1 for _ in enumerate(self.walk()))

    @property
    def path(self) -> str:
        return "/".join(reversed([shortlabel(node.data) for node in self.walk()]))

    @property
    def parent(self) -> Optional[NodeT]:
        return self._parent

    @parent.setter
    def parent(self, node: NodeT) -> None:
        if node is self:
            raise ValueError("Parent node cannot be itself.")
        self._parent = node

    @property
    def root(self) -> Optional[NodeT]:
        return self._root

    @root.setter
    def root(self, node: NodeT) -> None:
        if node is self:
            raise ValueError("Root node cannot be itself.")
        self._root = node



"""Compatibility utilities."""
from typing import IO, AnyStr

from .contexts import asyncnullcontext, nullcontext
from .futures import current_task
from .typing import AsyncContextManager, ChainMap, Counter, Deque, NoReturn

__all__ = [
    "AsyncContextManager",  # XXX moved to .typing
    "ChainMap",  # XXX moved to .typing
    "Counter",  # XXX moved to .typing
    "Deque",  # XXX moved to .typing
    "NoReturn",  # XXX moved to .typing
    "DummyContext",
    "OrderedDict",
    "want_bytes",
    "want_str",
    "isatty",
    "current_task",  # XXX moved to .futures
]

#: Dictionaries are ordered by default in Python 3.6
OrderedDict = dict


def want_bytes(s: AnyStr) -> bytes:
    """Convert string to bytes."""
    if isinstance(s, str):
        return s.encode()
    return s


def want_str(s: AnyStr) -> str:
    """Convert bytes to string."""
    if isinstance(s, bytes):
        return s.decode()
    return s


def isatty(fh: IO) -> bool:
    """Return True if fh has a controlling terminal.

    Notes:
        Use with e.g. :data:`sys.stdin`.
    """
    try:
        return fh.isatty()
    except AttributeError:
        return False


class DummyContext(nullcontext, asyncnullcontext):
    """Context for with-statement doing nothing."""

    # XXX deprecated, use nullcontext or asyncnullcontext



"""Context manager utilities."""
import typing
from types import TracebackType
from typing import Any, Type

if typing.TYPE_CHECKING:
    from ._py37_contextlib import (
        AbstractAsyncContextManager,
        AsyncExitStack,
        ExitStack,
        asynccontextmanager,
        nullcontext,
    )
else:
    try:  # pragma: no cover
        from contextlib import (
            AbstractAsyncContextManager,
            AsyncExitStack,
            ExitStack,
            asynccontextmanager,
            nullcontext,
        )
    except ImportError:  # pragma: no cover
        from ._py37_contextlib import (
            AbstractAsyncContextManager,
            AsyncExitStack,
            ExitStack,
            asynccontextmanager,
            nullcontext,
        )

__all__ = [
    "AbstractAsyncContextManager",
    "AsyncExitStack",
    "ExitStack",
    "asynccontextmanager",
    "nullcontext",
    "asyncnullcontext",
]


# Sphinx complains that stdlib is badly formatted :P

AsyncExitStack.__doc__ = """
Async context manager for dynamic management of a stack of exit
callbacks.

Example:
    >>> async with AsyncExitStack() as stack:
    ...    connections = [await stack.enter_async_context(get_connection())
    ...                   for i in range(5)]
    ...    # All opened connections will automatically be released at the
    ...    # end of the async with statement, even if attempts to open a
    ...    # connection later in the list raise an exception.
"""
asynccontextmanager.__doc__ = "asynccontextmanager decorator."
nullcontext.__doc__ = "Context that does nothing."


class asyncnullcontext(AbstractAsyncContextManager):
    """Context for async-with statement doing nothing."""

    enter_result: Any

    def __init__(self, enter_result: Any = None) -> None:
        self.enter_result = enter_result

    async def __aenter__(self) -> Any:
        return self.enter_result

    async def __aexit__(
        self,
        exc_type: Type[BaseException] = None,
        exc_val: BaseException = None,
        exc_tb: TracebackType = None,
    ) -> None:
        ...



import asyncio
import sys
from contextlib import asynccontextmanager
from functools import reduce
from itertools import chain
from typing import List, NamedTuple, Tuple
from unittest.mock import ANY, Mock, patch

if sys.version_info < (3, 8):
    from mock.mock import AsyncMock
else:
    from unittest.mock import AsyncMock

import pytest

from mode.timers import Timer
from mode.utils.aiter import aslice


@pytest.mark.asyncio
async def test_Timer_real_run():
    i = 0
    async for sleep_time in Timer(0.1, sleep=asyncio.sleep):
        assert sleep_time == pytest.approx(0.1, 2e-1)
        if i > 10:
            break
        i += 1


class Interval(NamedTuple):
    interval: float
    wakeup_time: float
    yield_time: float
    expected_new_interval: float


class test_Timer:

    # first clock value
    epoch = 9.0

    # the timer interval (how long we sleep between each iteration)
    interval = 1.0

    # how much we skew drifting intervals by.
    # early test will do (interval - skew)
    # late test will do (interval + skew)
    skew = 0.3

    # how long it takes to yield
    # Yield gives control back to the event loop, runs other tasks
    # and calls any callback associated with the timer, so this
    # time could be long in some cases.
    default_yield_s = 0.01

    @pytest.fixture()
    def clock(self):
        clock = Mock()
        clock.return_value = self.epoch
        return clock

    @pytest.fixture()
    def sleep(self):
        return AsyncMock()

    @pytest.fixture()
    def timer(self, *, clock, sleep) -> Timer:
        return Timer(
            self.interval,
            name="test",
            clock=clock,
            sleep=sleep,
        )

    @pytest.fixture()
    def first_interval(self):
        return self.new_interval()

    @pytest.mark.asyncio
    async def test_too_early(self, *, clock, timer, first_interval):
        interval = self.interval
        skew = self.skew
        intervals = [
            first_interval,  # 1st interval
            (None, None),  # 2nd interval
            (None, None),  # 3rd interval
            (interval - skew, None),  # 4th interval: sleep too short
            (None, interval + skew),  # 5th interval: overlaps
            (None, None),  # 6th interval
        ]
        async with self.assert_timer(timer, clock, intervals) as logger:
            logger.info.assert_called_once_with(
                "Timer %s woke up too early, with a drift "
                "of -%r runtime=%r sleeptime=%r",
                "test",
                ANY,
                ANY,
                ANY,
            )
            assert timer.drifting == 1
            assert timer.drifting_early == 1
            assert not timer.drifting_late

    @pytest.mark.asyncio
    async def test_too_late(self, *, clock, timer, first_interval):
        interval = self.interval
        skew = self.skew
        intervals = [
            first_interval,  # 1st interval
            (None, None),  # 2nd interval
            (None, None),  # 3rd interval
            (interval + skew, None),  # 4th interval: sleep too long
            (None, interval + skew),  # 5th interval: overlaps
            (None, None),  # 6th interval
        ]
        async with self.assert_timer(timer, clock, intervals) as logger:
            logger.info.assert_called_once_with(
                "Timer %s woke up too late, with a drift "
                "of +%r runtime=%r sleeptime=%r",
                "test",
                ANY,
                ANY,
                ANY,
            )
            assert timer.drifting == 1
            assert timer.drifting_late == 1
            assert not timer.drifting_early

    @asynccontextmanager
    async def assert_timer(self, timer, clock, interval_tuples):
        intervals = self.build_intervals(timer, *interval_tuples)
        print(intervals)
        clock_values = self.to_clock_values(*intervals)
        assert len(clock_values) == len(intervals) * 2
        clock.side_effect = clock_values

        with patch("mode.timers.logger") as logger:
            await self.assert_intervals(timer, intervals)
            yield logger

    def new_interval(
        self,
        interval: float = None,
        wakeup_time: float = None,
        yield_time: float = None,
        expected_new_interval: float = None,
    ) -> Interval:
        if interval is None:
            interval = self.interval
        if wakeup_time is None:
            wakeup_time = self.epoch + interval
        if yield_time is None:
            yield_time = wakeup_time + interval + self.default_yield_s
        if expected_new_interval is None:
            expected_new_interval = interval
        return Interval(
            interval,
            wakeup_time,
            yield_time,
            expected_new_interval,
        )

    def to_next_interval(
        self,
        timer: Timer,
        interval: Interval,
        sleep_time: float = None,
        yield_time: float = 0.1,
        expected_new_interval: float = None,
    ) -> Interval:
        if sleep_time is None:
            sleep_time = interval.interval + 0.001
        if yield_time is None:
            yield_time = self.default_yield_s
        next_wakeup_time = interval.yield_time + self.default_yield_s
        next_yield_time = next_wakeup_time + sleep_time
        time_spent_sleeping = interval.yield_time - interval.wakeup_time
        drift = interval.interval - time_spent_sleeping
        if expected_new_interval is None:
            expected_new_interval = timer.adjust_interval(drift)
        return Interval(
            interval=interval.interval,
            wakeup_time=next_wakeup_time,
            yield_time=next_yield_time,
            expected_new_interval=expected_new_interval,
        )

    def interval_to_clock_sequence(self, interval: Interval) -> List[float]:
        # Timer calls clock() twice per iteration,
        # so in an interval this provides the clock for wakeup time
        # and the yield time.
        return [interval.wakeup_time, interval.yield_time]

    def to_clock_values(self, *intervals: Interval) -> List[float]:
        return list(chain(*map(self.interval_to_clock_sequence, intervals)))

    def build_intervals(
        self, timer: Timer, first_interval: Interval, *values: Tuple[float, float]
    ) -> List[Interval]:
        """Build intervals from tuples of ``(sleep_time, yield_time)``.

        If a tuple is missing (is None), then default values
        will be taken from the previous interval.
        """

        intervals = [first_interval]

        def on_reduce(
            previous_interval: Interval, tup: Tuple[float, float]
        ) -> Interval:
            sleep_time, yield_time = tup
            next_interval = self.to_next_interval(
                timer,
                previous_interval,
                sleep_time=sleep_time,
                yield_time=yield_time,
            )
            intervals.append(next_interval)
            return next_interval

        reduce(on_reduce, values, first_interval)
        return intervals

    async def assert_intervals(self, timer: Timer, intervals: List[Interval]) -> None:
        assert await self.consume_timer(timer, limit=len(intervals)) == [
            interval.expected_new_interval for interval in intervals
        ]

    async def consume_timer(self, timer: Timer, limit: int) -> List[float]:
        return [sleep_time async for sleep_time in aslice(timer, 0, limit)]


class test_Timer_1s_half_second_skew(test_Timer):
    skew = 0.5


class test_Timer_30s_five_second_skew(test_Timer):
    interval = 30.0
    skew = 5.0


class test_Timer_30s_five_second_skew_late_epoch(test_Timer):
    epoch = 300000.0
    interval = 30.0
    skew = 5.0



# -*- coding: utf-8 -*-
"""AsyncIO Service-based programming."""
# :copyright: (c) 2017-2020, Robinhood Markets
#             (c) 2020-2022, faust-streaming Org
#             All rights reserved.
# :license:   BSD (3 Clause), see LICENSE for more details.
import re
import sys
from typing import NamedTuple

from .services import Service, task, timer  # noqa: E402
from .signals import BaseSignal, Signal, SyncSignal  # noqa: E402
from .supervisors import CrashingSupervisor  # noqa: E402
from .supervisors import (
    ForfeitOneForAllSupervisor,
    ForfeitOneForOneSupervisor,
    OneForAllSupervisor,
    OneForOneSupervisor,
    SupervisorStrategy,
)
from .types.services import ServiceT  # noqa: E402
from .types.signals import BaseSignalT, SignalT, SyncSignalT  # noqa: E402
from .types.supervisors import SupervisorStrategyT  # noqa: E402
from .utils.logging import flight_recorder, get_logger, setup_logging  # noqa: E402
from .utils.objects import label, shortlabel  # noqa: E402
from .utils.times import Seconds, want_seconds

if sys.version_info < (3, 8):
    from importlib_metadata import version
else:
    from importlib.metadata import version

__version__ = version("mode-streaming")
__author__ = "Faust Streaming"
__contact__ = "vpatki@wayfair.com, williambbarnhart@gmail.com"
__homepage__ = "https://github.com/faust-streaming/mode"
__docformat__ = "restructuredtext"

# -eof meta-


class version_info_t(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: str
    serial: str


# bumpversion can only search for {current_version}
# so we have to parse the version here.
_match = re.match(r"(\d+)\.(\d+).(\d+)(.+)?", __version__)
if _match is None:  # pragma: no cover
    raise RuntimeError("MODE VERSION HAS ILLEGAL FORMAT")
_temp = _match.groups()
VERSION = version_info = version_info_t(
    int(_temp[0]), int(_temp[1]), int(_temp[2]), _temp[3] or "", ""
)
del _match
del _temp
del re



"""Proxy to service.

Works like a service, but delegates to underlying service object.
"""
import abc
from typing import Any, ContextManager, Optional

from .services import ServiceBase
from .types import ServiceT
from .utils.types.trees import NodeT
from .utils.typing import AsyncContextManager

__all__ = ["ServiceProxy"]


class ServiceProxy(ServiceBase):
    """A service proxy delegates ServiceT methods to a composite service.

    Example:
        >>> class MyServiceProxy(ServiceProxy):
        ...
        ...     @cached_property
        ...     def _service(self) -> ServiceT:
        ...         return ActualService()

    Notes:
        This is used by Faust, and probably useful elsewhere!
        The Faust App is created at module-level, and it uses service
        proxy to ensure the event loop is not also created just by importing
        a module.
    """

    @property
    @abc.abstractmethod
    def _service(self) -> ServiceT:  # pragma: no cover
        ...

    def add_dependency(self, service: ServiceT) -> ServiceT:
        return self._service.add_dependency(service)

    async def add_runtime_dependency(self, service: ServiceT) -> ServiceT:
        return await self._service.add_runtime_dependency(service)

    async def add_async_context(self, context: AsyncContextManager) -> Any:
        return await self._service.add_async_context(context)

    def add_context(self, context: ContextManager) -> Any:
        return self._service.add_context(context)

    async def start(self) -> None:
        await self._service.start()

    async def maybe_start(self) -> bool:
        return await self._service.maybe_start()

    async def crash(self, reason: BaseException) -> None:
        await self._service.crash(reason)

    def _crash(self, reason: BaseException) -> None:
        self._service._crash(reason)

    async def stop(self) -> None:
        await self._service.stop()

    def service_reset(self) -> None:
        self._service.service_reset()

    async def restart(self) -> None:
        await self._service.restart()

    async def wait_until_stopped(self) -> None:
        await self._service.wait_until_stopped()

    def set_shutdown(self) -> None:
        self._service.set_shutdown()

    @property
    def started(self) -> bool:
        return self._service.started

    @property
    def crashed(self) -> bool:
        return self._service.crashed

    @property
    def should_stop(self) -> bool:
        return self._service.should_stop

    @property
    def state(self) -> str:
        return self._service.state

    @property
    def label(self) -> str:
        return type(self).__name__

    @property
    def shortlabel(self) -> str:
        return type(self).__name__

    @property
    def beacon(self) -> NodeT:
        return self._service.beacon

    @beacon.setter
    def beacon(self, beacon: NodeT) -> None:
        self._service.beacon = beacon

    @property
    def crash_reason(self) -> Optional[BaseException]:
        return self._service.crash_reason

    @crash_reason.setter
    def crash_reason(self, reason: Optional[BaseException]) -> None:
        self._service.crash_reason = reason



import asyncio
import signal
import sys
from contextlib import contextmanager
from signal import Signals
from unittest.mock import Mock, patch

if sys.version_info < (3, 8):
    from mock.mock import AsyncMock
else:
    from unittest.mock import AsyncMock

import pytest

from mode import Service
from mode.debug import BlockingDetector
from mode.utils.mocks import call, mask_module, patch_module
from mode.worker import Worker, exiting


def test_exiting():
    with patch("builtins.print") as print:
        with pytest.raises(SystemExit) as excinfo:
            with exiting(print_exception=True):
                raise KeyError()
        assert excinfo.value.code > 0
        print.assert_called()


class test_Worker:
    @pytest.fixture()
    def worker(self):
        return Worker(loglevel="INFO", logfile=None)

    def setup_method(self, method):
        self.setup_logging_patch = patch("mode.utils.logging.setup_logging")
        self.setup_logging = self.setup_logging_patch.start()

    def teardown_method(self):
        self.setup_logging_patch.stop()

    def test_constructor(self):
        stdout = Mock()
        stderr = Mock()
        assert Worker(stdout=stdout).stdout is stdout
        assert Worker(stderr=stderr).stderr is stderr
        assert Worker(stdout=None).stdout is not stdout
        assert Worker(stderr=None).stderr is not stderr

    def test_constructor_aligns_beacons(self):
        x = Service()
        y = Service()
        Worker(x, y)

    def test_say__quiet(self, worker):
        worker.quiet = True
        with patch("builtins.print") as print:
            worker.say("msg")
            print.assert_not_called()

    def test__say(self, worker):
        worker.quiet = False
        file = Mock()
        with patch("builtins.print") as print:
            worker._say("msg", file=file, foo=1)
            print.assert_called_once_with("msg", file=file, foo=1, end="\n")

    def test__say__default_file(self, worker):
        worker.quiet = False
        with patch("builtins.print") as print:
            worker._say("msg", file=None, end=".")
            print.assert_called_once_with("msg", file=worker.stdout, end=".")

    def test_on_init_dependencies(self, worker):
        workers = [Mock(), Mock(), Mock()]
        worker.services = workers
        assert worker.on_init_dependencies() == workers

    def _setup_for_on_first_start(self, worker):
        worker._setup_logging = Mock()
        worker.on_execute = AsyncMock()
        worker._add_monitor = AsyncMock()
        worker.install_signal_handlers = Mock()

    @pytest.mark.asyncio
    async def test_on_first_start(self, worker):
        self._setup_for_on_first_start(worker)
        worker.debug = False

        await worker.on_first_start()
        worker._setup_logging.assert_called_once_with()
        worker.on_execute.assert_awaited_once_with()
        worker._add_monitor.assert_not_called()
        worker.install_signal_handlers.assert_called_once_with()

        worker.debug = True
        await worker.on_first_start()
        worker._add_monitor.assert_awaited_once_with()

    @pytest.mark.asyncio
    async def test_on_first_start__override_logging(self):
        worker = Worker(override_logging=False)
        self._setup_for_on_first_start(worker)
        await worker.on_first_start()

        worker._setup_logging.assert_not_called()  # type: ignore[attr-defined]

    @pytest.mark.asyncio
    async def test_on_execute(self, worker):
        await worker.on_execute()

    @pytest.mark.parametrize(
        "loghandlers",
        [
            [],
            [Mock(), Mock()],
            [Mock()],
            None,
        ],
    )
    def test_setup_logging(self, loghandlers):
        worker_inst = Worker(
            loglevel=5,
            logfile="TEMP",
            loghandlers=loghandlers,
            logging_config=None,
        )
        worker_inst._setup_logging()
        self.setup_logging.assert_called_once_with(
            loglevel=5,
            logfile="TEMP",
            loghandlers=loghandlers or [],
            logging_config=None,
        )

    def test_setup_logging_raises_exception(self, worker):
        with patch("sys.stderr"):
            with patch("traceback.print_stack") as print_stack:
                with patch("mode.utils.logging.setup_logging") as sl:
                    sl.side_effect = KeyError("foo")
                    with pytest.raises(KeyError):
                        worker._setup_logging()

                    print_stack.side_effect = ValueError()
                    with pytest.raises(KeyError):
                        worker._setup_logging()

    def test_setup_logging__no_redirect(self, worker):
        worker.redirect_stdouts = False
        with patch("mode.utils.logging.setup_logging"):
            worker._setup_logging()

    @pytest.mark.asyncio
    async def test_maybe_start_blockdetection(self, worker):
        worker._blocking_detector = Mock(maybe_start=AsyncMock())
        worker.debug = False
        await worker.maybe_start_blockdetection()
        worker._blocking_detector.maybe_start.assert_not_awaited()

        worker.debug = True
        await worker.maybe_start_blockdetection()
        worker._blocking_detector.maybe_start.assert_awaited_once_with()

    def test_instal_signal_handlers(self, worker):
        worker._install_signal_handlers_windows = Mock()
        worker._install_signal_handlers_unix = Mock()

        worker.install_signal_handlers()
        if sys.platform == "win32":
            worker._install_signal_handlers_windows.assert_called_once_with()
        else:
            worker._install_signal_handlers_unix.assert_called_once_with()

    def test__install_signal_handlers_windows(self, worker):
        with patch("signal.signal") as sig:
            worker._install_signal_handlers_windows()
            sig.assert_called_once_with(signal.SIGTERM, worker._on_win_sigterm)

    @pytest.mark.skipif(sys.platform == "win32", reason="win32: no SIGUSR1")
    def test__install_signal_handlers_unix(self, worker):
        worker.loop = Mock()
        worker._install_signal_handlers_unix()
        worker.loop.add_signal_handler.assert_has_calls(
            [
                call(signal.SIGINT, worker._on_sigint),
                call(signal.SIGTERM, worker._on_sigterm),
                call(signal.SIGUSR1, worker._on_sigusr1),
            ]
        )

    def test__on_sigint(self, worker):
        worker._schedule_shutdown = Mock()
        worker._on_sigint()
        worker._schedule_shutdown.assert_called_once_with(signal.SIGINT)

    def test__on_sigterm(self, worker):
        worker._schedule_shutdown = Mock()
        worker._on_sigterm()
        worker._schedule_shutdown.assert_called_once_with(signal.SIGTERM)

    def test__on_win_sigterm(self, worker):
        worker._schedule_shutdown = Mock()
        worker._on_win_sigterm(1, Mock())
        worker._schedule_shutdown.assert_called_once_with(signal.SIGTERM)

    def test__on_sigusr1(self, worker):
        worker._cry = Mock()
        worker.add_future = Mock()
        worker._on_sigusr1()
        worker.add_future.assert_called_once_with(worker._cry.return_value)

    @pytest.mark.asyncio
    async def test__cry(self, worker):
        with patch("mode.utils.logging.cry") as cry:
            await worker._cry()
            cry.assert_called_once_with(file=worker.stderr)

    def test__schedule_shutdown(self, worker):
        with patch("asyncio.ensure_future") as ensure_future:
            worker._stop_on_signal = Mock()
            worker._schedule_shutdown(Signals.SIGTERM)
            assert worker._signal_stop_time
            ensure_future.assert_called_once_with(
                worker._stop_on_signal.return_value,
                loop=worker.loop,
            )
            worker._stop_on_signal.assert_called_once_with(Signals.SIGTERM)

            worker._schedule_shutdown(Signals.SIGTERM)

    @pytest.mark.asyncio
    async def test__stop_on_signal(self, worker):
        worker.stop = AsyncMock()
        await worker._stop_on_signal(Signals.SIGTERM)
        worker.stop.assert_awaited_once()

    def test_execute_from_commandline(self, worker):
        with self.patch_execute(worker) as ensure_future:
            with pytest.raises(SystemExit) as excinfo:
                worker.execute_from_commandline()
            assert excinfo.value.code == 0
            assert worker._starting_fut is ensure_future.return_value
            ensure_future.assert_called_once_with(
                worker.start.return_value, loop=worker.loop
            )
            worker.stop_and_shutdown.assert_called_once_with()

    def test_execute_from_commandline__MemoryError(self, worker):
        with self.patch_execute(worker):
            worker.start.side_effect = MemoryError()
            with pytest.raises(SystemExit) as excinfo:
                worker.execute_from_commandline()
            assert excinfo.value.code > 0

    def test_execute_from_commandline__CancelledError(self, worker):
        with self.patch_execute(worker):
            worker.start.side_effect = asyncio.CancelledError()
            with pytest.raises(SystemExit) as excinfo:
                worker.execute_from_commandline()
            assert excinfo.value.code == 0

    def test_execute_from_commandline__Exception(self, worker):
        with self.patch_execute(worker):
            worker.start.side_effect = KeyError("foo")
            with pytest.raises(SystemExit) as excinfo:
                worker.execute_from_commandline()
            assert excinfo.value.code > 0

    @contextmanager
    def patch_execute(self, worker):
        worker.loop = Mock()
        worker.start = Mock()
        worker.stop_and_shutdown = Mock()
        with patch("asyncio.ensure_future") as ensure_future:
            yield ensure_future

    def test_on_worker_shutdown(self, worker):
        worker.on_worker_shutdown()

    def test_stop_and_shutdown__stopping_worker(self, worker):
        worker.loop = Mock()
        worker.stop = Mock()
        worker._shutdown_loop = Mock()
        worker._signal_stop_future = None
        worker._stopped.clear()
        worker.loop = Mock()
        worker.stop_and_shutdown()

        worker.loop.run_until_complete.assert_called_with(worker.stop.return_value)

    def test__shutdown_loop(self, worker):
        with self.patch_shutdown_loop(worker, is_running=False):
            worker._shutdown_loop()

    def test__shutdown_loop__wait(self, worker):
        with self.patch_shutdown_loop(worker, is_running=True):

            def on_loop_stop():
                if worker.loop.stop.call_count >= 3:
                    worker.loop.is_running.return_value = False
                return worker.loop.stop.return_value

            worker.loop.stop.side_effect = on_loop_stop

            worker._shutdown_loop()

            assert worker.loop.stop.call_count == 4

    def test__shutdown_loop__wait_raises(self, worker):
        with self.patch_shutdown_loop(worker, is_running=True):
            worker.log.exception = Mock()

            def on_loop_stop():
                if worker.loop.stop.call_count >= 3:
                    print("MOO")
                    worker.loop.stop.side_effect = None
                    raise ValueError("moo")
                return worker.loop.stop.return_value

            worker.loop.stop.side_effect = on_loop_stop

            worker._shutdown_loop()

            assert worker.loop.stop.call_count == 4
            worker.log.exception.assert_called_once()

    def test__shutdown_loop__service_crashed(self, worker):
        worker.crash_reason = KeyError("foo")
        with self.patch_shutdown_loop(worker, is_running=False):
            with pytest.raises(KeyError):
                worker._shutdown_loop()

    @contextmanager
    def patch_shutdown_loop(self, worker, is_running=False):
        worker.loop = Mock()
        worker._gather_futures = Mock()
        worker._gather_all = Mock()
        worker.loop.is_running.return_value = is_running
        worker._sentinel_task = AsyncMock()
        with patch("asyncio.ensure_future") as ensure_future:
            with patch("asyncio.sleep", AsyncMock()):
                yield ensure_future

    @pytest.mark.asyncio
    async def test__sentinel_task(self, worker):
        with patch("asyncio.sleep", AsyncMock()) as sleep:
            await worker._sentinel_task()
            sleep.assert_called_once_with(1.0)

    def test__gather_all(self, worker):
        with patch("mode.worker.all_tasks") as all_tasks:
            with patch("asyncio.sleep", AsyncMock()):
                all_tasks.return_value = [Mock(), Mock(), Mock()]
                worker.loop = Mock()

                worker._gather_all()

                for task in all_tasks.return_value:
                    task.cancel.assert_called_once_with()

    def test__gather_all_early(self, worker):
        with patch("mode.worker.all_tasks") as all_tasks:
            with patch("asyncio.sleep"):
                worker.loop = Mock()

                def on_all_tasks(loop):
                    if all_tasks.call_count >= 5:
                        return []
                    return [Mock(), Mock(), Mock()]

                all_tasks.side_effect = on_all_tasks

                worker._gather_all()

                for task in all_tasks.return_value:
                    task.cancel.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_started(self, worker):
        worker.daemon = False
        worker.wait_until_stopped = AsyncMock()
        await worker.on_started()
        worker.wait_until_stopped.assert_not_awaited()
        worker.daemon = True
        await worker.on_started()
        worker.wait_until_stopped.assert_awaited_once()

    @pytest.mark.asyncio
    async def test__add_monitor(self, worker):
        worker.add_context = Mock()
        with patch_module("aiomonitor"):
            import aiomonitor

            await worker._add_monitor()

            worker.add_context.assert_called_once_with(
                aiomonitor.start_monitor.return_value
            )

            aiomonitor.start_monitor.assert_called_once_with(
                port=worker.console_port,
                loop=worker.loop,
            )

    @pytest.mark.asyncio
    async def test__add_monitor__no_aiomonitor(self, worker):
        worker.log.warning = Mock()
        with mask_module("aiomonitor"):
            await worker._add_monitor()
            worker.log.warning.assert_called_once()

    def test_repr_info(self, worker):
        assert repr(worker)

    def test_blocking_detector(self, worker):
        b = worker.blocking_detector
        assert isinstance(b, BlockingDetector)
        assert b.timeout == worker.blocking_timeout
        assert b.beacon.parent is worker.beacon
        assert b.loop is worker.loop



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
import sys

try:
    import platform

    _pyimp = platform.python_implementation
except (AttributeError, ImportError):

    def _pyimp():
        return "Python"


from setuptools import find_packages, setup

NAME = "mode-streaming"
EXTENSIONS = {"eventlet", "gevent", "uvloop"}
E_UNSUPPORTED_PYTHON = "%s 1.0 requires %%s %%s or later!" % (NAME,)  # noqa: S001

PYIMP = _pyimp()
if sys.version_info < (3, 6):
    raise Exception(E_UNSUPPORTED_PYTHON % (PYIMP, "3.6"))  # noqa: S001

from pathlib import Path  # noqa

README = Path("README.rst")

# -*- Classifiers -*-

classes = """
    Development Status :: 5 - Production/Stable
    License :: OSI Approved :: BSD License
    Programming Language :: Python :: 3 :: Only
    Programming Language :: Python :: 3.7
    Programming Language :: Python :: 3.8
    Programming Language :: Python :: 3.9
    Programming Language :: Python :: 3.10
    Programming Language :: Python :: Implementation :: CPython
    Operating System :: POSIX
    Operating System :: Microsoft :: Windows
    Operating System :: MacOS :: MacOS X
    Operating System :: Unix
    Environment :: No Input/Output (Daemon)
    Framework :: AsyncIO
    Intended Audience :: Developers
"""
classifiers = [s.strip() for s in classes.split("\n") if s]

# -*- Distribution Meta -*-

re_meta = re.compile(r"__(\w+?)__\s*=\s*(.*)")
re_doc = re.compile(r'^"""(.+?)"""')


def add_default(m):
    attr_name, attr_value = m.groups()
    return ((attr_name, attr_value.strip("\"'")),)


def add_doc(m):
    return (("doc", m.groups()[0]),)


pats = {re_meta: add_default, re_doc: add_doc}
here = Path(__file__).parent.absolute()
with open(here / "mode" / "__init__.py") as meta_fh:
    meta = {}
    for line in meta_fh:
        if line.strip() == "# -eof meta-":
            break
        for pattern, handler in pats.items():
            m = pattern.match(line.strip())
            if m:
                meta.update(handler(m))

# -*- Installation Requires -*-


def strip_comments(line):
    return line.split("#", 1)[0].strip()


def _pip_requirement(req):
    if req.startswith("-r "):
        _, path = req.split()
        return reqs(*path.split("/"))
    return [req]


def _reqs(*f):
    path = (Path.cwd() / "requirements").joinpath(*f)
    reqs = (strip_comments(line) for line in path.open().readlines())
    return [_pip_requirement(r) for r in reqs if r]


def reqs(*f):
    return [req for subreq in _reqs(*f) for req in subreq]


def extras(*p):
    """Parse requirement in the requirements/extras/ directory."""
    return reqs("extras", *p)


def extras_require():
    """Get map of all extra requirements."""
    return {x: extras(x + ".txt") for x in EXTENSIONS}


# -*- Long Description -*-


if README.exists():
    long_description = README.read_text(encoding="utf-8")
else:
    long_description = "See http://pypi.org/project/{}".format(NAME)

# -*- %%% -*-

packages = find_packages(
    exclude=["t", "t.*", "docs", "docs.*", "examples", "examples.*"],
)
assert not any(package.startswith("t.") for package in packages)


setup(
    name=NAME,
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description=meta["doc"],
    author=meta["author"],
    author_email=meta["contact"],
    url=meta["homepage"],
    platforms=["any"],
    license="BSD",
    keywords="asyncio service bootsteps graph coroutine",
    packages=packages,
    include_package_data=True,
    # PEP-561: https://www.python.org/dev/peps/pep-0561/
    package_data={"mode": ["py.typed"]},
    zip_safe=False,
    install_requires=reqs("default.txt"),
    tests_require=reqs("test.txt"),
    extras_require=extras_require(),
    python_requires="~=3.6",
    classifiers=classifiers,
    long_description=long_description,
    long_description_content_type="text/x-rst",
)



"""Backport of :mod:`typing` additions in Python 3.7."""
# pragma: no cover
import typing

__all__ = [
    "AsyncContextManager",
    "AsyncGenerator",
    "ChainMap",
    "Counter",
    "Deque",
    "NoReturn",
    "Protocol",
]

if typing.TYPE_CHECKING:
    from typing import AsyncContextManager
else:
    try:
        from typing import AsyncContextManager  # noqa: F811
    except ImportError:
        from typing_extensions import AsyncContextManager

if typing.TYPE_CHECKING:
    from typing import AsyncGenerator
else:
    try:
        from typing import AsyncGenerator
    except ImportError:  # Python 3.6.0
        from typing_extensions import AsyncGenerator


if typing.TYPE_CHECKING:
    from typing import ChainMap
else:
    try:
        from typing import ChainMap  # noqa: F811
    except ImportError:
        from typing_extensions import ChainMap


if typing.TYPE_CHECKING:
    from typing import Counter
else:
    try:
        from typing import Counter  # noqa: F811
    except ImportError:
        from typing_extensions import Counter


if typing.TYPE_CHECKING:
    from typing import Deque
else:
    try:
        from typing import Deque  # noqa: F811
    except ImportError:
        from typing_extensions import Deque


if typing.TYPE_CHECKING:
    from typing import NoReturn
else:
    try:
        from typing import NoReturn  # noqa: F811
    except ImportError:
        from typing_extensions import NoReturn


if typing.TYPE_CHECKING:
    from typing import Protocol
else:
    try:
        from typing import Protocol  # noqa: F811
    except ImportError:
        from typing_extensions import Protocol



"""Time, date and timezone related utilities."""
import abc
import asyncio
import sys
import time
from datetime import timedelta
from functools import singledispatch
from types import TracebackType
from typing import Callable, List, Mapping, NamedTuple, Optional, Type, Union

from .text import pluralize
from .typing import AsyncContextManager

__all__ = [
    "Bucket",
    "Seconds",
    "TokenBucket",
    "rate",
    "rate_limit",
    "want_seconds",
    "humanize_seconds",
    "humanize_seconds_ago",
]

TIME_MONOTONIC: Callable[[], float]
if sys.platform == "win32":
    TIME_MONOTONIC = time.time
else:
    TIME_MONOTONIC = time.monotonic

#: Seconds can be expressed as float or :class:`~datetime.timedelta`,
Seconds = Union[timedelta, float, str]


class Unit(NamedTuple):
    name: str
    value: float
    format: Callable[[float], str]  # noqa: E701


TIME_UNITS: List[Unit] = [
    Unit("day", 60 * 60 * 24.0, lambda n: format(n, ".2f")),
    Unit("hour", 60 * 60.0, lambda n: format(n, ".2f")),
    Unit("minute", 60.0, lambda n: format(n, ".2f")),
    Unit("second", 1.0, lambda n: format(n, ".2f")),
]

#: What the characters in a "rate" string means.
#: E.g. 8/s is "eight in one second"
RATE_MODIFIER_MAP: Mapping[str, Callable[[float], float]] = {
    "s": lambda n: n,
    "m": lambda n: n / 60.0,
    "h": lambda n: n / 60.0 / 60.0,
    "d": lambda n: n / 60.0 / 60.0 / 24,
}


class Bucket(AsyncContextManager):
    """Rate limiting state.

    A bucket "pours" tokens at a rate of ``rate`` per second (or over').

    Calling `bucket.pour()`, pours one token by default, and returns
    :const:`True` if that amount can be poured now, or :const:`False` if the
    caller has to wait.

    If this returns :const:`False`, it's prudent to either sleep or raise
    an exception::

        if not bucket.pour():
            await asyncio.sleep(bucket.expected_time())

    If you want to consume multiple tokens in one go then specify the number::

        if not bucket.pour(10):
            await asyncio.sleep(bucket.expected_time(10))

    This class can also be used as an async. context manager, but in that case
    can only consume one tokens at a time::

        async with bucket:
            # do something

    By default the async. context manager will suspend the current coroutine
    and sleep until as soon as the time that a token can be consumed.

    If you wish you can also raise an exception, instead of sleeping, by
    providing the ``raises`` keyword argument::

        # hundred tokens in one second, and async with: raises TimeoutError

        class MyError(Exception):
            pass

        bucket = Bucket(100, over=1.0, raises=MyError)

        async with bucket:
            # do something

    """

    rate: float
    capacity: float

    _tokens: float

    def __init__(
        self,
        rate: Seconds,
        over: Seconds = 1.0,
        *,
        fill_rate: Seconds = None,
        capacity: Seconds = None,
        raises: Type[BaseException] = None,
        loop: asyncio.AbstractEventLoop = None
    ) -> None:
        self.rate = want_seconds(rate)
        self.capacity = want_seconds(over)
        self.raises = raises
        self.loop = loop
        self._tokens = self.capacity
        self.__post_init__()

    def __post_init__(self) -> None:
        ...

    @abc.abstractmethod
    def pour(self, tokens: int = 1) -> bool:
        ...

    @abc.abstractmethod
    def expected_time(self, tokens: int = 1) -> float:
        ...

    @property
    @abc.abstractmethod
    def tokens(self) -> float:
        ...

    @property
    def fill_rate(self) -> float:
        #: Defaults to rate! If you want the bucket to fill up
        #: faster/slower, then just override this.
        return self.rate

    async def __aenter__(self) -> "Bucket":
        if not self.pour():
            if self.raises:
                raise self.raises()
            expected_time = self.expected_time()
            await asyncio.sleep(expected_time)
        return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException] = None,
        exc_val: BaseException = None,
        exc_tb: TracebackType = None,
    ) -> Optional[bool]:
        return None


class TokenBucket(Bucket):
    """Rate limiting using the token bucket algorithm."""

    _tokens: float
    _last_pour: float

    def __post_init__(self) -> None:
        self._last_pour = TIME_MONOTONIC()

    def pour(self, tokens: int = 1) -> bool:
        need = tokens
        have = self.tokens
        if have < need:
            return False
        self._tokens -= tokens
        return True

    def expected_time(self, tokens: int = 1) -> float:
        have = self._tokens
        need = max(tokens, have)
        time_left = (need - have) / self.fill_rate
        return max(time_left, 0.0)

    @property
    def tokens(self) -> float:
        now = TIME_MONOTONIC()
        if now < self._last_pour:
            return self._tokens
        if self._tokens < self.capacity:
            delta = self.fill_rate * (now - self._last_pour)
            self._tokens = min(self.capacity, self._tokens + delta)
            self._last_pour = now
        return self._tokens


@singledispatch
def rate(r: float) -> float:
    """Convert rate string (`"100/m"`, `"2/h"` or `"0.5/s"`) to seconds."""
    return r


@rate.register(str)
def _rate_str(r: str) -> float:  # noqa: F811
    ops, _, modifier = r.partition("/")
    return RATE_MODIFIER_MAP[modifier or "s"](float(ops)) or 0


@rate.register(int)  # noqa: F811
def _rate_int(r: int) -> float:
    return float(r)


@rate.register(type(None))  # noqa: F811
def _rate_None(r: None) -> float:
    return 0.0


def rate_limit(
    rate: float,
    over: Seconds = 1.0,
    *,
    bucket_type: Type[Bucket] = TokenBucket,
    raises: Type[BaseException] = None,
    loop: asyncio.AbstractEventLoop = None
) -> Bucket:
    """Create rate limiting manager."""
    return bucket_type(rate, over, raises=raises, loop=loop)


@singledispatch
def want_seconds(s: float) -> float:
    """Convert :data:`Seconds` to float."""
    return s


@want_seconds.register(str)  # noqa: F811
def _want_seconds_str(s: str) -> float:
    return rate(s)


@want_seconds.register(timedelta)  # noqa: F811
def _want_seconds_timedelta(s: timedelta) -> float:
    return s.total_seconds()


def humanize_seconds(
    secs: float,
    *,
    prefix: str = "",
    suffix: str = "",
    sep: str = "",
    now: str = "now",
    microseconds: bool = False
) -> str:
    """Show seconds in human form.

    For example, 60 becomes "1 minute", and 7200 becomes "2 hours".

    Arguments:
        secs: Seconds to format (as :class:`float` or :class:`int`).
        prefix (str): can be used to add a preposition to the output
            (e.g., 'in' will give 'in 1 second', but add nothing to 'now').
        suffix (str): same as prefix, adds suffix unless 'now'.
        sep (str): separator between prefix and number.
        now (str): Literal 'now'.
        microseconds (bool): Include microseconds.
    """
    secs = float(format(float(secs), ".2f"))
    for unit, divider, formatter in TIME_UNITS:
        if secs >= divider:
            w = secs / float(divider)
            return "{0}{1}{2} {3}{4}".format(
                prefix, sep, formatter(w), pluralize(int(w), unit), suffix
            )
    if microseconds and secs > 0.0:
        return "{prefix}{sep}{0:.2f} seconds{suffix}".format(
            secs, sep=sep, prefix=prefix, suffix=suffix
        )
    return now


def humanize_seconds_ago(
    secs: float,
    *,
    prefix: str = "",
    suffix: str = " ago",
    sep: str = "",
    now: str = "just now",
    microseconds: bool = False
) -> str:
    """Show seconds in "3.33 seconds ago" form.

    If seconds are less than one, returns "just now".
    """
    return humanize_seconds(
        secs,
        prefix=prefix,
        suffix=suffix,
        sep=sep,
        now=now,
        microseconds=microseconds,
    )



"""Mocking and testing utilities."""
import builtins
import sys
import types
import unittest.mock
from contextlib import contextmanager
from types import ModuleType
from typing import Any, Iterator, cast
from unittest.mock import MagicMock

__all__ = ["IN", "call", "mask_module", "patch_module"]


class IN:
    """Class used to check for multiple alternatives.

    .. code-block:: python

        assert foo.value IN(a, b)

    """

    def __init__(self, *alternatives: Any) -> None:
        self.alternatives = alternatives

    def __eq__(self, other: Any) -> bool:
        return other in self.alternatives

    def __ne__(self, other: Any) -> bool:
        return other not in self.alternatives

    def __repr__(self) -> str:
        sep = " | "
        return f"<IN: {sep.join(map(str, self.alternatives))}>"


@contextmanager
def patch_module(*names: str, new_callable: Any = MagicMock) -> Iterator:
    """Mock one or modules such that every attribute is a :class:`Mock`."""
    prev = {}

    class MockModule(types.ModuleType):
        def __getattr__(self, attr: str) -> Any:
            setattr(self, attr, new_callable())
            return types.ModuleType.__getattribute__(self, attr)

    mods = []
    for name in names:
        try:
            prev[name] = sys.modules[name]
        except KeyError:
            pass
        mod = sys.modules[name] = MockModule(name)
        mods.append(mod)
    try:
        yield mods
    finally:
        for name in names:
            try:
                sys.modules[name] = prev[name]
            except KeyError:
                try:
                    del sys.modules[name]
                except KeyError:
                    pass


@contextmanager
def mask_module(*modnames: str) -> Iterator:
    """Ban some modules from being importable inside the context.

    For example::

        >>> with mask_module('sys'):
        ...     try:
        ...         import sys
        ...     except ImportError:
        ...         print('sys not found')
        sys not found

        >>> import sys  # noqa
        >>> sys.version
        (2, 5, 2, 'final', 0)

    Taken from
    http://bitbucket.org/runeh/snippets/src/tip/missing_modules.py

    """
    realimport = builtins.__import__

    def myimp(name: str, *args: Any, **kwargs: Any) -> ModuleType:
        if name in modnames:
            raise ImportError(f"No module named {name}")
        else:
            return cast(ModuleType, realimport(name, *args, **kwargs))

    builtins.__import__ = myimp
    try:
        yield
    finally:
        builtins.__import__ = realimport


class _Call(unittest.mock._Call):
    # Fixes bug in Python 3.8 where call is an instance of unittest.mock._Call
    # but call.__doc__ returns a mocked method and not the class attribute.

    def __getattr__(self, attr: str) -> Any:
        if attr == "__doc__":
            return unittest.mock._Call.__doc__
        return super().__getattr__(attr)


call = _Call(from_kall=False)
